'''
Definitions of the kernels.

Note:
    Outside the kernels' support, the function return undefined values! If
    arbitrary values of radius are use, it has to be taken care of those outside
    the support manually.
                                                                              r
    Furthermore the arguments are always normed by the smoothing length: u = --- .
                                                                             2 h

Examples and doctests:
    >>> cubic(0.0)
    2.5464790894703255
    >>> cubic(0.5)
    0.6366197723675
    >>> cubic(1.0)
    0.0
    >>> Wendland_C4_vec(np.linspace(0,1,12))
    array([4.92385605e+00, 4.56340351e+00, 3.65811505e+00, 2.55311070e+00,
           1.54488080e+00, 7.95912455e-01, 3.36297380e-01, 1.08637670e-01,
           2.33707743e-02, 2.44034893e-03, 4.47381991e-05, 0.00000000e+00])

    Test for the equality of all the vector versions and for normation:
    >>> for name in sorted(kernels):
    ...     kernel = kernels[name]
    ...     for vec_kernels in [vector_kernels]:
    ...         if name not in vec_kernels:
    ...             continue
    ...         c_kernel = getattr(C.cpygad,name.replace(' ','_'),None)
    ...         if c_kernel:
    ...             print('found C version of "%s" kernel' % name)
    ...         kernel_vec = vec_kernels[name]
    ...         u = np.linspace(0,1,100)
    ...         w = kernel_vec(u)
    ...         for uu,ww in zip(u,w):
    ...             if not np.all(np.abs(ww - kernel(uu)) < 1e-6):
    ...                 print(name, 'not equal to its vector version')
    ...                 break
    ...             if c_kernel:
    ...                 if not np.all(np.abs(ww - c_kernel(uu,1.0)) < 1e-3):
    ...                     print(ww, c_kernel(uu,1.0))
    ...                     print(name, 'not equal to its C version')
    ...                     break
    ...         I = np.sum(4*np.pi*u**2 * w) / (len(w)-1)
    ...         if not abs(I-1.0) < 1e-6:
    ...             print(name, 'is not normed to 1, but to', I)
    found C version of "Wendland C2" kernel
    found C version of "Wendland C4" kernel
    found C version of "Wendland C6" kernel
    found C version of "cubic" kernel
    found C version of "quartic" kernel
    found C version of "quintic" kernel
'''
__all__ = [
    'kernels', 'vector_kernels',
    'cubic', 'cubic_vec',
    'quartic', 'quartic_vec',
    'quintic', 'quintic_vec',
    'Wendland_C2', 'Wendland_C2_vec',
    'Wendland_C4', 'Wendland_C4_vec',
    'Wendland_C6', 'Wendland_C6_vec',
]

import numpy as np
from .. import C


def cubic(u):
    '''
    The cubic kernel:

        8   /   2           \               1
        -- | 6 u (u - 1) + 1 |      if  u < -
        pi  \               /               2

          16  /     \ 3
        - -- | 1 - u |              otherwise
          pi  \     /

    where u = r/(2h) <= 1.
    '''
    if u < 0.5:
        return 2.5464790894703255 * (1. + 6. * (u - 1.) * u ** 2)
    else:
        return 5.09295817894 * (1. - u) ** 3


def cubic_vec(u):
    '''The vector version of the cubic kernel.'''
    u = np.asarray(u)
    w = np.empty(len(u))
    mask = u < 0.5
    u_masked = u[mask]
    w[mask] = 2.5464790894703255 * (1. + 6. * (u_masked - 1.) * u_masked ** 2)
    mask = ~mask
    w[mask] = 5.09295817894 * (1. - u[mask]) ** 3
    return w


def quartic(u):
    '''
    The quartic kernel:

         15625  / /    \ 4     / 2    \ 4     / 1    \ 4 \               1
        ------ | |1 - u | - 5 |  - - u | + 10 | - - u |   |      if  u < -
        512 pi  \ \    /       \ 3    /       \ 3    /   /               5

         15625  / /    \ 4     / 2    \ 4 \                          1        3
        ------ | |1 - u | - 5 |  - - u | + |                     if  - <= u < -
        512 pi  \ \    /       \ 3    /   /                          5        5

         15625    /    \ 4
        ------   |1 - u |                                        otherwise
        512 pi    \    /

    where u = r/(2h) <= 1.
    '''
    if u < 0.2:
        return 9.71404681957369 * ((1. - u) ** 4 - 5. * (0.6 - u) ** 4 + 10. * (0.2 - u) ** 4)
    elif u < 0.6:
        return 9.71404681957369 * ((1. - u) ** 4 - 5. * (0.6 - u) ** 4)
    else:
        return 9.71404681957369 * ((1. - u) ** 4)


def quartic_vec(u):
    '''The vector version of the quartic kernel.'''
    u = np.asarray(u)
    w = (1. - u) ** 4
    mask = u < 0.6
    u_masked = u[mask]
    w[mask] -= 5. * (0.6 - u_masked) ** 4
    mask = u < 0.2
    u_masked = u[mask]
    w[mask] += 10. * (0.2 - u_masked) ** 4
    return 9.71404681957369 * w


def quintic(u):
    '''
    The quintic kernel:

         2187  / /    \ 5     / 2    \ 5     / 1    \ 5 \               1
        ----- | |1 - u | - 6 |  - - u | + 15 | - - u |   |      if  u < -
        40 pi  \ \    /       \ 3    /       \ 3    /   /               3

         2187  / /    \ 5     / 2    \ 5 \                          1        2
        ----- | |1 - u | - 6 |  - - u | + |                     if  - <= u < -
        40 pi  \ \    /       \ 3    /   /                          3        3

         2187    /    \ 5
        -----   |1 - u |                                        otherwise
        40 pi    \    /

    where u = r/(2h) <= 1.
    '''
    if u < 0.3333333333333333:
        return 17.403593027098754 * (
                    (1. - u) ** 5 - 6. * (0.6666666666666667 - u) ** 5 + 15. * (0.3333333333333333 - u) ** 5)
    elif u < 0.6666666666666667:
        return 17.403593027098754 * ((1. - u) ** 5 - 6. * (0.6666666666666667 - u) ** 5)
    else:
        return 17.403593027098754 * ((1. - u) ** 5)


def quintic_vec(u):
    '''The vector version of the quintic kernel.'''
    u = np.asarray(u)
    w = (1. - u) ** 5
    mask = u < 0.6666666666666667
    u_masked = u[mask]
    w[mask] -= 6. * (0.6666666666666667 - u_masked) ** 5
    mask = u < 0.3333333333333333
    u_masked = u[mask]
    w[mask] += 15. * (0.3333333333333333 - u_masked) ** 5
    return 17.403593027098754 * w


def Wendland_C2(u):
    '''
    The Wendland C2 kernel:

                  4 /       \
        21 (1 - u) |         |
        ---------- | 1 + 4 u |
           2 pi     \       /

    where u = r/(2h) <= 1.
    '''
    return 3.3422538049298023 * (1. - u) ** 4 * (1. + 4.0 * u)


def Wendland_C2_vec(u):
    '''The vector version of the Wendland C2 kernel.'''
    u = np.asarray(u)
    return 3.3422538049298023 * (1. - u) ** 4 * (1. + 4.0 * u)


def Wendland_C4(u):
    '''
    The Wendland C4 kernel:

                   6  /    2          \
        495 (1 - u)  | 35 u            |
        ------------ | ----- + 6 u + 1 |
           32 pi      \  3            /

    where u = r/(2h) <= 1.
    '''
    return 4.923856051905513 * (1. - u) ** 6 * (1. + (6. + 11.6666666667 * u) * u)


def Wendland_C4_vec(u):
    '''The vector version of the Wendland C4 kernel.'''
    u = np.asarray(u)
    return 4.923856051905513 * ((1. - u) ** 6 * (1. + (6. + 11.6666666667 * u) * u))


def Wendland_C6(u):
    '''
    The Wendland C6 kernel:

                    8
        1365 (1 - u)   /    3      2         \
        ------------- | 32 u + 25 u + 8 u + 1 |
            64 pi      \                     /

    where u = r/(2h) <= 1.
    '''
    return 6.78895304126366 * (1. - u) ** 8 * (1. + (8. + (25. + 32. * u) * u) * u)


def Wendland_C6_vec(u):
    '''The vector version of the Wendland C6 kernel.'''
    u = np.asarray(u)
    return 6.78895304126366 * ((1. - u) ** 8 * (1. + (8. + (25. + 32. * u) * u) * u))


kernels = {
    'cubic': cubic,
    'quartic': quartic,
    'quintic': quintic,
    'Wendland C2': Wendland_C2,
    'Wendland C4': Wendland_C4,
    'Wendland C6': Wendland_C6,
}
vector_kernels = {
    'cubic': cubic_vec,
    'quartic': quartic_vec,
    'quintic': quintic_vec,
    'Wendland C2': Wendland_C2_vec,
    'Wendland C4': Wendland_C4_vec,
    'Wendland C6': Wendland_C6_vec,
}





'''
High level access for (single) Gadget files.

Doctesting is implicity done in the snapshot module.
'''
__all__ = ['FileReader', 'write']

import os
from .. import environment
from ..environment import secure_get_h5py
h5py = secure_get_h5py()
from .lowlevel_file import *
from ..units import *
from . import config
import numpy as np
import warnings
import sys

class FileReader(object):
    '''
    A handler for reading Gadget files of all formats.

    Args:
        filename (str):         The path to the Gadget file.
        unclear_blocks (str):   What to do the blocks for which the block info is
                                unclear (cannot be infered). Possible modes are:
                                * exception:    raise an IOError
                                * warning:      print a warning to the stderr
                                * ignore:       guess what
                                If it is None, the value from the `gadget.cfg` is
                                taken.

    Raises:
        RuntimeError:       If the information could not be infered or if the
                            given dtype of the given family is unknown.
    '''
    def __init__(self, filename, unclear_blocks=None):
        if not os.path.exists(os.path.expanduser(filename)):
            raise IOError('"%s" does not exist!' % filename)
        self._filename = filename
        if unclear_blocks is None:
            unclear_blocks = config.general['unclear_blocks']
        if h5py.is_hdf5(self._filename):
            self._format, self._endianness = 3, None
            with h5py.File(self._filename, 'r') as gfile:
                self._header = read_header(gfile, self._format,
                                           self._endianness)
                self._info = get_block_info(gfile, self._format,
                                            self._endianness, self._header,
                                            unclear_blocks)
        else:
            with open(os.path.expanduser(self._filename), 'rb') as gfile:
                self._format, self._endianness = \
                            get_format_and_endianness(gfile)
                self._header = read_header(gfile, self._format,
                                           self._endianness)
                self._info = get_block_info(gfile, self._format,
                                            self._endianness, self._header,
                                            unclear_blocks)

    @property
    def filename(self):
        '''The (basic) filename.'''
        return self._filename

    @property
    def format(self):
        '''The Gadget file format.'''
        return self._format

    @property
    def endianness(self):
        '''The endianess of the file.'''
        return self._endianness

    @property
    def header(self):
        '''The file header in form of a dictionary.'''
        return self._header

    def infos(self):
        '''Return a sorted list of the block informations.'''
        try:
            erg = sorted(list(self._info.values()), key=lambda e: e.start_pos)
        except Exception as e:
            erg = list(self._info.values())
        return erg

    def block_names(self):
        '''Return a list of all available blocks.'''
        return list(self._info.keys())

    def has_block(self, name):
        '''Check whether there is a block with the asked name.'''
        return name in self._info

    def read_block(self, name, gad_units=None, unit=None):
        '''
        Args:
            name (str):         The name of the block. For HDF5 files: If not the
                                actual block name is passed, it is tried to
                                convert from the standard Gadget block name to the
                                HDF5 one via 'gadget.config.std_name_to_HDF5'.
            gad_units (dict):   The basic gadget units. If None, take the units
                                defined in the `gadget.cfg`.
            unit (Unit):        The unit to give the data. If it is None, use a
                                default one if defined, Unit(1) otherwise.

        Returns:
            block (UnitArr):    The loaded block.
        '''
        block = self._info[name]
        N = sum(self._header['N_part'][pt] for pt in range(6)
                if block.ptypes[pt])

        if self._format == 3:
            with h5py.File(self._filename, 'r') as gfile:
                HDF5_name = config.std_name_to_HDF5.get(name,name)
                data = np.empty(N*block.dimension, dtype=block.dtype)
                if block.dimension > 1:
                    data = data.reshape( (N,block.dimension) )
                off = 0
                for pt in range(6):
                    if block.ptypes[pt]:
                        ds = gfile['PartType%d' % pt][HDF5_name]
                        data[off:off+ds.shape[0]] = ds[:]   # actually load data
                        off += len(ds)
        else:
            filename = os.path.expanduser(self._filename)
            with open(filename, 'r') as gfile:
                gfile.seek(block.start_pos)
                data = np.fromfile(gfile,
                                   dtype=self._endianness+block.type_descr,
                                   count=N*block.dimension)
                if block.dimension > 1:
                    data = data.reshape( (N,block.dimension) )

        if block.name == 'MASS':
            pos = 0
            for pt in range(6):
                if self._header['mass'][pt]:
                    data = np.insert(data,
                                     pos,
                                     np.ones(self._header['N_part'][pt],
                                             dtype=block.dtype) * \
                                        self._header['mass'][pt])
                    """
                    for md, mh in [(data[pos], self._header['mass'][pt]),
                                   (data[pos+self._header['N_part'][pt]-1],
                                        self._header['mass'][pt])]:
                        assert (md-mh)/mh < 1e-6
                    """
                pos += self._header['N_part'][pt]

        data = data.view(UnitArr)
        if unit is not None:
            data.units = unit
        else:
            try:
                data.units = config.get_block_units(block.name,gad_units)
            except KeyError:
                data.units = 1
            except:
                raise

        return data

def _create_header(s, gad_units=None, double_prec=None):
    '''
    Create a header dictionary as needed by the low-level file routines.

    Up to now: Just one file and no mass section...
    TODO: change that!?

    Args:
        s (Snap):           The (sub-)snapshot to create the header from.
        gad_units (dict):   The Gadget unit system to use. If None, use the one
                            stored in the snapshot.
        double_prec (bool): The value for the flag for double precision. If it is
                            None, the value from the snapshot properties is taken.

    Returns:
        header (dict):  The header dictionary (including the unused section).
    '''
    header = dict()

    if gad_units is None:
        gad_units = s.gadget_units
    else:
        tmp, gad_units = gad_units, s.gadget_units.copy()
        gad_units.update( tmp )

    header['N_part']        = s.parts
    header['mass']          = [0.0]*6
    header['time']          = s.time
    header['redshift']      = s.redshift
    header['flg_sfr']       = int(bool(s.properties['flg_sfr']))
    header['flg_feedback']  = int(bool(s.properties['flg_feedback']))
    header['N_part_all']    = s.parts
    header['flg_cool']      = int(bool(s.properties['flg_cool']))
    header['N_files']       = 1
    header['boxsize']       = float(s.boxsize.in_units_of(gad_units['LENGTH']))
    header['Omega_m']       = s.cosmology.Omega_m
    header['Omega_Lambda']  = s.cosmology.Omega_Lambda
    header['h_0']           = s.cosmology.h_0
    header['flg_age']       = int(bool(s.properties['flg_age']))
    header['flg_metals']    = int(bool(s.properties['flg_metals']))
    header['flg_entropy_instead_u'] = int(bool(
                                        s.properties['flg_entropy_instead_u']))
    if double_prec is None:
        header['flg_doubleprecision'] = s.properties['flg_doubleprecision']
    else:
        header['flg_doubleprecision'] = int(bool(double_prec))
    header['flg_ic_info']   = int(bool(s.properties['flg_ic_info']))
    header['lpt_scalingfactor'] = \
            0.0 if s.properties['lpt_scalingfactor'] is None \
            else float(s.properties['lpt_scalingfactor'])
    header['unused']        = ' '*68

    return header

def write(snap, filename, blocks=None, gformat=2, endianness='native',
          infoblock=True, double_prec=False, gad_units=None, overwrite=False):
    '''
    Write a snapshot.

    TODO:
        * implement putting masses to header if they are all equal for a particle
          type
        * implement writing to multiple files (needed?)

    Args:
        snap (Snap):        The (sub-)snapshot to write.
        filename (str):     The path to write to.
        blocks (list):      A list of the names of the blocks to write (as in the
                            attributes, not the Gadget names, i.e. 'pos' rather
                            than 'POS '). The blocks are written in the order
                            given in this list. The standard blocks 'pos', 'vel',
                            'ID', and 'mass, however, are always written and
                            always in this order at the beginning.
                            If None, all loadable blocks (but no derived ones) of
                            the snapshots are written. The order is random, except
                            that the first blocks are those specified in
                            gadget.config.block_order (non-available blocks are
                            skipped).
        gformat (int):      The file format (either 1, 2, or 3 (HDF5)).
        endianness (str):   The endianess ('='/'native', '<'/'little', '>'/'big').
        infoblock (bool):   Whether to write an infoblock.
        double_prec (bool): Whether to write in double precision.
        gad_units (dict):   The Gadget unit system to use. If None, use the one
                            stored in the snapshot.
        overwrite (bool):   Allow to overwrite, if the file already exists.
    '''
    if os.path.exists(os.path.expanduser(filename)) and not overwrite:
        raise IOError('The file "%s" already exists. ' % filename +
                      'Consider "overwrite=True"!')
    if blocks is None:
        blocks = snap.loadable_blocks()
        tmp = []
        for name in config.block_order:
            for attrname, lname in snap._root._load_name.items():
                if name == lname:
                    name = attrname
                    break
            if name in blocks:
                blocks.remove(name)
                tmp.append(name)
        blocks = tmp + blocks
    else:
        std_blocks = ['pos', 'vel', 'ID', 'mass']
        blocks = std_blocks + [name for name in blocks if name not in std_blocks]
    if gformat not in [1,2,3]:
        raise ValueError('Unknow format %s!' % gformat)
    endianness = {'native':'=',
                  'little':'<',
                  'big':'>'}.get(endianness,endianness)
    if endianness not in ['=','<','>']:
        raise ValueError('Unknown endianness %s!' % endianness)
    if gad_units is None:
        gad_units = snap.gadget_units
    else:
        tmp, gad_units = gad_units, snap.gadget_units.copy()
        gad_units.update( tmp )

    # prepare
    header = _create_header(snap, gad_units=gad_units, double_prec=double_prec)
    data = {}
    info = {}
    for name in blocks:
        # do not use block hosts, since we want to be able to write sub-snapshots,
        # as well
        for sub in [snap]+[getattr(snap,fam) for fam in config.families]:
            if len(sub)>0 and name in sub.available_blocks():
                ptypes = [bool(N) for N in sub.parts]
                if name in data:
                    # only update, if we have the bigger block
                    if sum(ptypes) <= sum(info[name].ptypes):
                        continue
                block = sub[name]
                size = block.dtype.itemsize*np.prod(block.shape)
                dtype = block.dtype.base
                if dtype.kind == 'f':
                    if double_prec and dtype != 'float64':
                        size = 8 * size / dtype.itemsize
                        dtype = np.dtype('float64')
                    elif not double_prec and dtype != 'float32':
                        size = 4 * size / dtype.itemsize
                        dtype = np.dtype('float32')
                block_name = snap._root._load_name.get(name, '%-4s'%name.upper())
                info[name] = BlockInfo(name=block_name, dtype=dtype,
                                       dimension=np.prod(block.shape[1:]),
                                       ptypes=ptypes, start_pos=None, size=size)
                if block_name in config.block_units:
                    units = config.get_block_units(block_name, gad_units)
                else:
                    units = sub[name].units
                    warnings.warn('Blocks that do not have default units, are '
                                  'stored in their current units -- other blocks '
                                  'are converted!')
                if gformat == 3:
                    data[name] = [None] * 6
                    for pt in range(6):
                        if sub._N_part[pt]:
                            data[name][pt] = sub[[pt]][name] \
                                                .astype(info[name].dtype) \
                                                .in_units_of(units, subs=snap)
                else:
                    data[name] = sub[name] \
                                    .astype(info[name].dtype) \
                                    .in_units_of(units, subs=snap)

    # it might happen, that a sub-snapshot shall be written for which not all
    # blocks are actually present (e.g. it does not contain gas though the root
    # does)
    for name in set(blocks)-set(data.keys()):
        print('WARNING: block "%s" is not present for ' % name + \
                             'this (sub-)snapshot and, hence, not written!', file=sys.stderr)
        blocks.remove(name)

    if gformat == 3:
        with h5py.File(filename, 'w') as gfile:
            write_header(gfile, header, gformat, endianness)
            for pt in range(6):
                gfile.create_group('PartType%d' % pt)

            for name in blocks:
                d = data[name]
                for pt in range(6):
                    if d[pt] is not None:
                        units = d[pt].units
                hdf5name = snap._root._load_name.get(name)
                if hdf5name in config.std_name_to_HDF5:
                    hdf5name = config.std_name_to_HDF5[hdf5name]
                if hdf5name is None:
                    hdf5name = name

                if environment.verbose >= environment.VERBOSE_NORMAL:
                    print('writing block', hdf5name, end=' ')
                    print('(dtype=%s, units=%s)...' % (info[name].dtype, units), end=' ')
                    sys.stdout.flush()
                for pt in range(6):
                    if not snap._root._block_avail[name][pt]:
                        continue
                    group = gfile['PartType%d' % pt]
                    gf_block = group.create_dataset(
                            hdf5name,
                            d[pt].shape,
                            d[pt].dtype)
                    gf_block[...] = d[pt]
                if environment.verbose >= environment.VERBOSE_NORMAL:
                    print('done.')
                    sys.stdout.flush()
    else:
        with open(os.path.expanduser(filename), 'wb') as gfile:
            write_header(gfile, header, gformat, endianness)
            for name in blocks:
                block_name = info[name].name
                info[name].start_pos = gfile.tell()+4
                if gformat == 2:
                    info[name].start_pos += 4+8+4
                if environment.verbose >= environment.VERBOSE_NORMAL:
                    print('writing block', block_name, end=' ')
                    print('(dtype=%s, units=%s)...' % (info[name].dtype,
                            data[name].units), end=' ')
                    sys.stdout.flush()
                sys.stdout.flush()
                write_block(gfile, block_name, data[name], gformat, endianness)
                if environment.verbose >= environment.VERBOSE_NORMAL:
                    print('done.')
                    sys.stdout.flush()
            info = sorted(list(info.values()), key=lambda x: x.start_pos)
            write_info(gfile, info, gformat, endianness)



'''
A module to deal with Gadget snapshots. Amongst other objects, it also defines a
class Snap to hold a snapshot.

Example:
    Make sure that the Gadget snapshot used here exists and the destination
    snapshot file does not yet exist. This example copies the snapshot by first
    reading it in and then writing it do the destination file (amongst other
    things). Finally it is tested whether the copy is equivalent to the source
    in the sense that if it is read all blocks are the same. Later on some basic
    sub-snapshots are created and the concept of derived arrays is briefly
    demonstrated.

    >>> snap_tmp_file = 'test.gdt'
    >>> s = Snapshot(module_dir+'snaps/AqA_ICs_C02_200_nm.gdt', physical=False,
    ...          gad_units={'LENGTH':'cMpc/h_0'})
    >>> s
    <Snap "AqA_ICs_C02_200_nm.gdt.0-8"; N=2,951,686; z=127.000>
    >>> s['pos'].units
    load block pos... done.
    Unit("cMpc h_0**-1")
    >>> s.gadget_units
    {'LENGTH': 'cMpc/h_0', 'VELOCITY': 'a**(1/2) km / s', 'MASS': '1e10 Msol / h_0'}
    >>> print('current age of the universe: %s' % s.cosmic_time().in_units_of('Myr'))
    current age of the universe: 12.869734401309785 [Myr]
    >>> s.loadable_blocks()
    ['pos', 'vel', 'ID', 'mass']
    >>> if not set(s.deriveable_blocks()) >= set('Epot vx jzjc vy vz rcyl momentum angmom E dV vrad Ekin temp vcirc r jcirc y x z'.split()):
    ...     print(' '.join(sorted(s.deriveable_blocks())))
    >>> assert set(s.all_blocks()) == set(s.loadable_blocks() + s.deriveable_blocks())
    >>> mwgt_pos = np.tensordot(s['mass'], s['pos'], axes=1).view(UnitArr)
    load block mass... done.
    >>> mwgt_pos.units = s['mass'].units * s['pos'].units
    >>> com = mwgt_pos / s['mass'].sum()
    >>> np.linalg.norm(com - UnitArr([50.2]*3, 'cMpc/h_0')) < 0.5
    True

    And the physical distance between the center of mass and the unweighted mean
    of the positions is:
    (Conversion from 'ckpc/h_0' to 'kpc' is done automatically: the values for 'a'
    and 'h_0' are taken from the associated snapshot and substitued.)
    >>> np.sqrt(np.sum( (com - s['pos'].mean(axis=0))**2 )).in_units_of('kpc', subs=s) < 10.0
    UnitArr(True)

    Whereas the physical dimensions of the simulation's box are:
    >>> s.boxsize
    SimArr(100.0, units="cMpc h_0**-1", snap="AqA_ICs_C02_200_nm.gdt.0-8")
    >>> s['pos'].max(axis=0) - s['pos'].min(axis=0)
    UnitArr([97.78572, 97.81533, 97.83026], dtype=float32, units="cMpc h_0**-1")

    >>> s.load_all_blocks()
    load block vel... done.
    load block ID... done.

    It is also possible to slice entire snapshot, e.g. to access single families
    of the snapshot (gas, stars, dm, bh, baryons) or to mask them with a
    np.ndarray of bools (for more information see SubSnap).
    In fact, in order to access blocks that are only for certains families, one has
    to restrict the snapshot to appropiately.
    >>> s = Snapshot(module_dir+'snaps/snap_M1196_4x_470', physical=False)
    >>> s.gas['rho']
    load block rho... done.
    SimArr([2.1393036e-03, 8.1749255e-04, 9.4836479e-04, ..., 4.6801919e-08,
            3.9165805e-08, 3.4469060e-08],
           dtype=float32, units="1e+10 Msol ckpc**-3 h_0**2", snap="snap_M1196_4x_470":gas)

    The derived block R (radius) is updated automatically, if the block pos it
    depends on changes:
    >>> s['r']
    load block pos... done.
    derive block r... done.
    SimArr([59671.72500919, 59671.40716648, 59671.54775578, ...,
            59672.52310468, 59671.98484173, 59671.74536916],
           units="ckpc h_0**-1", snap="snap_M1196_4x_470")
    >>> assert np.all( s['r'] == dist(s['pos']) )
    >>> s['pos'] -= UnitArr([34622.7,35522.7,33180.5], 'ckpc/h_0')
    >>> s['r']
    derive block r... done.
    SimArr([ 9.41927424,  9.69844804,  9.6864571 , ..., 25.08852488,
            23.95322772, 24.6890346 ],
           units="ckpc h_0**-1", snap="snap_M1196_4x_470")
    >>> s.cache_derived = False
    >>> s.delete_blocks(derived=True)
    >>> s['r'].units
    derive block r... done.
    Unit("ckpc h_0**-1")
    >>> s['r'].units    # now gets re-derived
    derive block r... done.
    Unit("ckpc h_0**-1")
    >>> s.stars['age'].units    # this block is always cached! (see `derived.cfg`)
    load block form_time... done.
    derive block age... done.
    Unit("Gyr")
    >>> s.stars['age'].units
    Unit("Gyr")
    >>> s.cache_derived = True

    >>> s.to_physical_units()
    >>> sub = s[s['r'] < UnitScalar('30 kpc')]
    derive block r... done.
    >>> sub
    <Snap "snap_M1196_4x_470":masked; N=121,013; z=0.000>
    >>> sub.gas
    <Snap "snap_M1196_4x_470":masked:gas; N=11,191; z=0.000>

    # You can also get (almost) arbitrary combinations of the blocks
    >>> s.gas.get('dist(pos)**2 / rho')
    SimArr([1.54323637e-05, 4.28144806e-05, 3.68149753e-05, ...,
            2.25019852e+04, 2.70823020e+04, 3.08840657e+04],
           units="Msol**-1 kpc**5", snap="snap_M1196_4x_470":gas)
    >>> assert np.max(np.abs( (s.get('dist(pos)') - s['r']) / s['r'] )) < 1e-6
    >>> del s['pos']
    >>> s['r']
    SimArr([13.08232505, 13.47006664, 13.45341258, ..., 34.84517185,
            33.268371  , 34.2903246 ], units="kpc", snap="snap_M1196_4x_470")
    >>> s['pos']
    load block pos... done.
    SimArr([[48074.324, 49335.855, 46081.395],
            [48074.023, 49335.785, 46080.992],
            [48073.973, 49335.945, 46081.223],
            ...,
            [48067.195, 49359.18 , 46065.844],
            [48067.496, 49357.258, 46066.246],
            [48066.184, 49357.805, 46066.43 ]],
           dtype=float32, units="kpc", snap="snap_M1196_4x_470")
    >>> s['r']
    derive block r... done.
    SimArr([82877.39261021, 82876.95257089, 82877.14659861, ...,
            82878.50049942, 82877.75400349, 82877.42052568],
           units="kpc", snap="snap_M1196_4x_470")

    One can test for available blocks and families by the 'in' opertator:
    >>> 'r' in s
    True
    >>> if not set(s.available_blocks()) >= set('Epot pot pos jzjc vx vy vz rcyl momentum mass vel angmom jcirc E vrad ID Ekin vcirc r y x z'.split()):
    ...     print(' '.join(sorted(s.available_blocks())))
    >>> s.delete_blocks(derived=True)
    >>> 'r' in s
    True
    >>> 'stars' in s
    True

    New custom blocks can be set easily (though have to fit the (sub-)snapshot):
    >>> sub = s.baryons[s.baryons['metallicity']>1e-3].stars
    load block Z... done.
    derive block elements... done.
    derive block H... done.
    derive block He... done.
    derive block metals... done.
    derive block metallicity... done.
    >>> s.stars['new'] = np.ones(len(s.stars))
    >>> sub['new']
    SimArr([1., 1., 1., ..., 1., 1., 1.], snap="snap_M1196_4x_470":stars)

    >>> dest_file = module_dir+'test.gad'
    >>> assert not os.path.exists(dest_file)
    >>> gadget.write(sub, dest_file, blocks=['pos', 'ID', 'r'])
    load block vel... done.
    load block ID... done.
    load block mass... done.
    derive block r... done.
    writing block POS  (dtype=float32, units=[ckpc h_0**-1])... done.
    writing block VEL  (dtype=float32, units=[a**1/2 km s**-1])... done.
    writing block ID   (dtype=uint32, units=[1])... done.
    writing block MASS (dtype=float32, units=[1e+10 Msol h_0**-1])... done.
    writing block R    (dtype=float32, units=[kpc])... done.
    >>> sub_copy = Snapshot(dest_file, physical=True)
    >>> sub_copy['r'].units = 'kpc' # notice: block r gets *loaded*!
    ...                             # the units are those from before!
    ...                             # and are not in the config!
    load block r... done.
    >>> assert sub.parts == sub_copy.parts
    >>> assert np.max(np.abs((sub['pos'] - sub_copy['pos']) / sub['pos'])) < 1e-6
    load block pos... done.
    >>> assert np.max(np.abs((sub['r'] - sub_copy['r']) / sub['r'])) < 1e-6
    >>> assert np.abs((sub.boxsize - sub_copy.boxsize) / sub.boxsize) < 1e-6
    >>> import os
    >>> os.remove(dest_file)
    >>> del sub, sub_copy
    >>> s = Snapshot(module_dir+'snaps/snap_M1196_4x_470', physical=False)
    >>> gadget.write(s, dest_file)  # doctest:+ELLIPSIS
    load block pos... done.
    load block vel... done.
    load block ID... done.
    load block mass... done.
    load block u... done.
    ...
    writing block POS  (dtype=float32, units=[ckpc h_0**-1])... done.
    writing block VEL  (dtype=float32, units=[a**1/2 km s**-1])... done.
    writing block ID   (dtype=uint32, units=[1])... done.
    writing block MASS (dtype=float32, units=[1e+10 Msol h_0**-1])... done.
    writing block U    (dtype=float32, units=[a**2 km**2 s**-2])... done.
    ...
    >>> s2 = Snapshot(dest_file, physical=False)
    >>> s2.load_all_blocks()    # doctest:+ELLIPSIS
    load block ... done.
    ...
    >>> for name in s.loadable_blocks():
    ...     b  = s.get_host_subsnap(name)[name]
    ...     b2 = s2.get_host_subsnap(name)[name]
    ...     err = np.max( np.abs((b - b2) / b) )
    ...     if err[np.isfinite(err)] > 1e-6:
    ...         print(name, b, b2)
    >>> os.remove(dest_file)

    >>> dest_file_hdf5 = module_dir+'test.hdf5'
    >>> assert not os.path.exists(dest_file_hdf5)
    >>> gadget.write(s, dest_file_hdf5) # doctest:+ELLIPSIS
    ...
    writing block POS  (dtype=float32, units=[ckpc h_0**-1])... done.
    writing block VEL  (dtype=float32, units=[a**1/2 km s**-1])... done.
    writing block ID   (dtype=uint32, units=[1])... done.
    writing block MASS (dtype=float32, units=[1e+10 Msol h_0**-1])... done.
    writing block U    (dtype=float32, units=[a**2 km**2 s**-2])... done.
    ...
    >>> tmp = Snapshot(dest_file_hdf5, physical=False)
    >>> gadget.write(tmp, dest_file)    # doctest:+ELLIPSIS
    load block pos... done.
    load block vel... done.
    load block ID... done.
    load block mass... done.
    load block u... done.
    ...
    writing block POS  (dtype=float32, units=[ckpc h_0**-1])... done.
    writing block VEL  (dtype=float32, units=[a**1/2 km s**-1])... done.
    writing block ID   (dtype=uint32, units=[1])... done.
    writing block MASS (dtype=float32, units=[1e+10 Msol h_0**-1])... done.
    writing block U    (dtype=float32, units=[a**2 km**2 s**-2])... done.
    ...
    >>> del tmp
    >>> os.remove(dest_file_hdf5)
    >>> s2 = Snapshot(dest_file, physical=False)
    >>> for name in s.loadable_blocks():    # doctest:+ELLIPSIS
    ...     b  = s.get_host_subsnap(name)[name]
    ...     b2 = s2.get_host_subsnap(name)[name]
    ...     assert np.all( b == b2 )
    load block ...
    >>> assert s.redshift == s2.redshift
    >>> for name, prop in s.properties.items():
    ...     if not prop == s2.properties[name]:
    ...         print(name, prop, s2.properties[name])
    >>> os.remove(dest_file)
    >>> del s2

    Some basic testing for (reading) HDF5 snapshots:
    >>> s = Snapshot(module_dir+'snaps/snap_M1196_4x_conv_470.hdf5', physical=False)
    >>> s
    <Snap "snap_M1196_4x_conv_470.hdf5"; N=2,079,055; z=0.000>
    >>> s.cosmology
    FLRWCosmo(h_0=0.72, O_Lambda=0.74, O_m=0.26, O_b=0.0416, sigma_8=None, n_s=None)
    >>> s['pos']
    load block pos... done.
    SimArr([[34613.516, 35521.816, 33178.605],
            [34613.297, 35521.766, 33178.316],
            [34613.26 , 35521.883, 33178.48 ],
            ...,
            [34608.383, 35538.61 , 33167.41 ],
            [34608.598, 35537.227, 33167.7  ],
            [34607.652, 35537.62 , 33167.832]],
           dtype=float32, units="ckpc h_0**-1", snap="snap_M1196_4x_conv_470.hdf5")
    >>> s['r']
    derive block r... done.
    SimArr([59671.72500919, 59671.40716648, 59671.54775578, ...,
            59672.52310468, 59671.98484173, 59671.74536916],
           units="ckpc h_0**-1", snap="snap_M1196_4x_conv_470.hdf5")
    >>> s.parts
    [921708, 1001472, 56796, 19315, 79764, 0]
    >>> sub = s.SubSnap([0,1,2,4,5])
    >>> sub
    <Snap "snap_M1196_4x_conv_470.hdf5":pts=[0,1,2,4,5]; N=2,059,740; z=0.000>
    >>> sub.parts
    [921708, 1001472, 56796, 0, 79764, 0]
    >>> s.SubSnap([0,2,4]).parts
    [921708, 0, 56796, 0, 79764, 0]
'''
__all__ = ['write', 'Snapshot', 'Snap', 'SubSnapshot']

import sys
import os.path
from .. import gadget
from ..gadget import write
from .. import physics
from .. import utils
from ..units import *
from ..utils import dist
from .. import environment
from ..environment import module_dir
import numpy as np
import warnings
import ast
import weakref
from . import derived
import fnmatch


class Snapshot(object):
    '''
    A class holding a Gadget snapshot.

    Args:
        physical (bool):    Whether to convert to physical units on loading.
        cosmological (bool):Whether the simulation was a cosmological one.
        gad_units (dict):   Alternative base units (LENGTH, VELOCITY, MASS) for
                            this snapshot.
    '''

    def _initsnap(self, filename, base, suffix, physical=False, load_double_prec=False, cosmological=None,
                  gad_units=None, unclear_blocks=None, H_neutral_only=None):
        '''
        Create a snapshot from file (without loading the blocks, yet).

        Args:
            filename (str):         The path to the snapshot. If it is distributed
                                    over several files, you shall omit the trailing
                                    (of inbetween in case of an HDF5 file) '.0'.
            physical (bool):        Whether to convert to physical units on loading.
            load_double_prec (bool):Force to load all blocks in double precision.
                                    Equivalent with setting the snapshots attribute.
            cosmological (bool):    Explicitly tell if the simulation is a
                                    cosmological one.
            gad_units (dict):       Alternative base units (LENGTH, VELOCITY, MASS)
                                    for this snapshot. The default base units units
                                    are updated, meaning one can also just change one
                                    them.
            unclear_blocks (str):   What to do the blocks for which the block info is
                                    unclear (cannot be infered). Possible modes are:
                                    * exception:    raise an IOError
                                    * warning:      print a warning to the stderr
                                    * ignore:       guess what
                                    If it is None, the value from the `gadget.cfg` is
                                    taken.

        Raises:
            IOError:            If the snapshot does not exist.
            RuntimeError:       If the information could not be infered or if the
                                given dtype of the given family is unknown.
        '''

        from .sim_arr import SimArr
        s = self
        s._filename = os.path.abspath(base + suffix)
        s._descriptor = os.path.basename(base) + suffix
        # need first (maybe only) file for basic information
        greader = gadget.FileReader(filename, unclear_blocks=unclear_blocks)
        s._file_handlers = [greader]

        s._block_avail = {
            block.name: block.ptypes
            for block in greader.infos()
            if block.dtype is not None}

        s._N_part = list(map(int, greader.header['N_part_all']))
        s._time = greader.header['time']
        s._redshift = greader.header['redshift']
        s._boxsize = SimArr(greader.header['boxsize'],
                            units=s._gad_units['LENGTH'],
                            snap=s)
        s._cosmology = physics.FLRWCosmo(  # Note: Omega_b at default!
            h_0=greader.header['h_0'],
            Omega_Lambda=greader.header['Omega_Lambda'],
            Omega_m=greader.header['Omega_m'])
        s._properties = {k: v for k, v in greader.header.items()
                         if k not in ['N_part', 'mass', 'time', 'redshift', 'N_part_all',
                                      'N_files', 'h_0', 'Omega_Lambda', 'Omega_m', 'boxsize',
                                      'unused']}
        if s._cosmological is None:
            s._cosmological = abs(s.scale_factor - s.time) < 1e-6
        s._load_double_prec = bool(load_double_prec)

        if physical:
            s._boxsize.convert_to(s._boxsize.units.free_of_factors(['a', 'h_0']),
                                  subs=s)

        if greader.header['N_files'] > 1:
            s._descriptor += '.0-' + str(greader.header['N_files'])
            # ensure Python int's to avoid overflows
            N_part = list(map(int, greader.header['N_part']))
            for n in range(1, greader.header['N_files']):  # first already done
                filename = base + '.' + str(n) + suffix
                greader = gadget.FileReader(filename, unclear_blocks=unclear_blocks)
                s._file_handlers.append(greader)
                # ensure Python int's to avoid overflows
                for i in range(6): N_part[i] += int(greader.header['N_part'][i])
                # update loadable blocks:
                for block in greader.infos():
                    if block.name in s._block_avail:
                        s._block_avail[block.name] = [(o or n) for o, n \
                                                      in zip(s._block_avail[block.name], block.ptypes)]
                    else:
                        s._block_avail[block.name] = block.ptypes
            if N_part != s._N_part:
                # more particles than fit into a native int
                s._N_part = N_part

        # Process block names: make standard names lower case (except ID) and replace
        # spaces with underscores for HDF5 names. Also strip names
        s._load_name = {}
        for name, block in list(s._block_avail.items()):
            if s._file_handlers[0]._format == 3 \
                    and '%-4s' % name not in gadget.std_name_to_HDF5:
                new_name = name.strip()
            else:
                new_name = name.strip().lower()

            # some renaming
            if new_name in ['id', 'z']:
                new_name = new_name.upper()
            elif new_name == 'age':
                new_name = 'form_time'

            s._load_name[new_name] = name
            s._block_avail[new_name] = s._block_avail[name]
            if name != new_name:
                if environment.verbose >= environment.VERBOSE_TALKY \
                        and new_name.lower() != name.strip().lower():
                    print('renamed block "%s" to %s' % (name, new_name))
                del s._block_avail[name]  # blocks should not appear twice
        # now the mass block is named 'mass' for all cases (HDF5 or other)
        s._block_avail['mass'] = [n > 0 for n in s._N_part]

        s.fill_derived_rules()

        s._descriptor = '"' + s._descriptor + '"'

        return s

    def __init__(self, filename, physical=False, load_double_prec=False, cosmological=None,
                                 gad_units=None, unclear_blocks=None, H_neutral_only=None):

        filename = os.path.expandvars(filename)
        filename = os.path.expanduser(filename)
        # handle different filenames, e.g. cases where the snapshot is distributed
        # over severale files
        base, suffix = os.path.splitext(filename)
        if suffix != '.hdf5':
            base, suffix = filename, ''
        if not os.path.exists(filename):
            filename = base + '.0' + suffix
            if not os.path.exists(filename):
                raise IOError('Snapshot "%s%s" does not exist!' % (base, suffix))

        self._filename              = '<none>'
        self._descriptor            = 'new'
        self._file_handlers         = []
        self._gad_units             = gadget.default_gadget_units.copy()
        if gad_units:
            self._gad_units.update( gad_units )
        self._load_name             = {}    # from attribute name to actual block
                                            # name as to use for loading
        self._block_avail           = {}    # keys are lower case, stripped names
                                            # or the hdf5 names with underscores
        self._derive_rule_deps      = {}
        self._blocks                = {}
        self._cosmological          = cosmological
        self._N_part                = [0] * 6
        self._time                  = 1.0
        self._redshift              = 0.0
        self._boxsize               = 0.0
        self._cosmology             = physics.Planck2013()
        self._properties            = {}
        self._load_double_prec      = False
        self._phys_units_requested  = bool(physical)
        self._trans_at_load         = []
        self._root                  = self
        self._base                  = None
        self._cache_derived         = derived.general['cache_derived']
        self._always_cache          = set() # _derive_rule_deps is empty; gets
                                            # filled in Snap() with
                                            # derived.general['always_cache']
        # Horst: copy configuration to snapshot-property
        if H_neutral_only is None:
            self._H_neutral_only        = gadget.config.general['H_neutral_only']
        else:
            self._H_neutral_only        = H_neutral_only

        # Actual initialization is done in the factory function Snap. Just do some
        # basic setting of the attributes to ensure that even snapshot created by
        # just _Snap are somewhat functioning.
        self._initsnap(filename, base, suffix, physical=physical, load_double_prec=load_double_prec, cosmological=cosmological,
                  gad_units=gad_units, unclear_blocks=unclear_blocks, H_neutral_only=H_neutral_only)

    @property
    def filename(self):
        '''The absolut path of the snapshot. For multiple files, omitting the
        numbers in the filename, so that it can be reused for loading.'''
        return self._root._filename

    @property
    def descriptor(self):
        '''A short description of the origin of the (sub-)snapshot.'''
        return self._descriptor

    @property
    def gadget_units(self):
        return self._root._gad_units.copy()

    @property
    def cosmological(self):
        '''Is the simulation cosmological?'''
        return self._root._cosmological

    @property
    def N_files(self):
        '''The number of file the snapshot is distributed over.'''
        return len(self._root._file_handlers)

    # Horst: new property-method to query neutral-H setting
    @property
    def H_neutral_only(self):
        '''The configuration property for neutral H.'''
        return self._root._H_neutral_only

    def __len__(self):
        return sum(self._N_part)

    @property
    def parts(self):
        '''A list of the number of particles of each particle type in the
        snapshot.'''
        return self._N_part[:]

    @property
    def time(self):
        '''The time variable of the snapshot.'''
        return self._root._time

    @property
    def redshift(self):
        '''The redshift of the snapshot.'''
        return self._root._redshift

    @property
    def boxsize(self):
        '''The side length of the simulation box.'''
        return self._root._boxsize

    @property
    def cosmology(self):
        '''
        A copy of the simulations cosmology as taken from the header.

        The baryon fraction is the default (of the matter fraction) as
        physics.FLRWCosmo yields.
        '''
        return self._root._cosmology.copy()

    @property
    def properties(self):
        '''A copy of some more properties from the header(s).'''
        return self._root._properties.copy()

    @property
    def cache_derived(self):
        '''Whether to cache derived blocks (or to calculate them every time).'''
        return self._root._cache_derived

    @cache_derived.setter
    def cache_derived(self, sd):
        self._root._cache_derived = bool(sd)

    @property
    def always_cached_derived(self):
        '''List of blocks that are always changed.'''
        return list(self._root._always_cache)

    @property
    def load_double_prec(self):
        '''Whether to convert all blocks with floating point data to double
        precision on load.'''
        return self._root._load_double_prec

    @load_double_prec.setter
    def load_double_prec(self, do):
        self._root._load_double_prec = bool(do)

    @property
    def phys_units_requested(self):
        '''Whether to convert all blocks to physical units on load.'''
        return self._root._phys_units_requested

    @phys_units_requested.setter
    def phys_units_requested(self, do):
        self._root._phys_units_requested = bool(do)

    @property
    def root(self):
        '''The underlying snapshot of this (sub-)snapshot (self for root
        itself).'''
        return self._root

    @property
    def base(self):
        '''The snapshot this one is sliced/masked from (None for root).'''
        return self._base

    @property
    def scale_factor(self):
        '''
        The scale factor of the universe at the time of the snapshot.

        This is *not* simply the time, but calculated from the redshift, in order
        to also be correct in the case of non-cosmological simulations.
        '''
        return physics.z2a(self.redshift)

    def SubSnap(self, mask):
        '''
        A factory function for creating masked/sliced sub-snapshots.

        Sub-snapshots should always be instantiated with this function or the bracket
        notation (or FamilySubSnap).

        Args:
            base (Snap):    The snapshot to create the sub-snapshot from.

            mask:           The mask to use. It can be:

                            * slice:
                                Slice the entire snapshot as one would do with
                                np.ndarray's. Blocks are then slices of the base
                                snapshot. It is taken care of those which are now
                                available, but were not for the base.

                            * np.ndarray[bool]:
                                Has to have the length of the snapshot. The
                                sub-snapshot then consists of all the particles for
                                which the entry is True. Otherwise the same as for a
                                slice.

                            * particle types (list):
                                Create a sub-snapshot of the specified particle types
                                only.

                            * mask class (SnapMask):
                                Create a sub-snapshot according to the mask.

                            * mask class (Halo):
                                Create a sub-snapshot according to the mask of the
                                halo (an IDMask).

                            * index list (np.ndarray[int]):
                                This can in fact also be a tuple with one element,
                                which is such a index list, as returned by np.where.
                                This is similar to passing a boolean mask, but here
                                the indices (not the IDs) are passed explicitly.

                                With contrast to passing this directly to the block
                                arrays, the particles are *not reordered* here and
                                every particle can only *occur once*, since in the
                                back, this converts the mask to a boolean one with the
                                use of sets. I did this for a combination of little
                                effort in writing code (reusing) and speed (using
                                sets).
                                TODO:
                                    - Change this behaviour?
                                    - Really keep this option?

        Returns:
            sub (_SubSnap):     The sub-snapshot.

        Raises:
            KeyError:           If the mask was not understood.
        '''
        from .masks import SnapMask
        from ..analysis.halo import Halo

        base = self

        if isinstance(mask, slice) \
                or (isinstance(mask, np.ndarray) and mask.dtype == bool):
            # slices and masks are handled directly by _SubSnap
            return SubSnapshot(base, mask)

        elif isinstance(mask, list):
            ptypes = sorted(set(mask))
            # precalculating N_part is faster than the standard way in _SubSnap
            N_part = [(base._N_part[pt] if pt in ptypes else 0) for pt in range(6)]
            if utils.is_consecutive(ptypes):
                # slicing is faster than masking!
                sub = slice(sum(base._N_part[:ptypes[0]]),
                            sum(base._N_part[:ptypes[-1] + 1]))
            else:
                l = [(np.ones(base._N_part[pt], bool) if pt in ptypes
                      else np.zeros(base._N_part[pt], bool)) for pt in range(6)]
                sub = np.concatenate(l)
            sub = SubSnapshot(base, sub, N_part)
            sub._descriptor = base._descriptor + ':pts=' + str(ptypes).replace(' ', '')
            return sub

        elif isinstance(mask, SnapMask) or isinstance(mask, Halo):
            if isinstance(mask, Halo):
                mask = mask.mask
            sub = SubSnapshot(base, mask.get_mask_for(base))
            sub._descriptor = base._descriptor + ':' + str(mask)
            return sub

        elif isinstance(mask, np.ndarray) and mask.dtype.kind == 'i':
            # I probably do not want to keep this. Increases backward compability,
            # though.
            warnings.warn('Consider using the faster boolean masks!')
            # is convering into boolean array the best choice?
            warnings.warn('Indexed snapshot masking does not reorder!')
            idx_set = set(mask)
            if len(idx_set) < len(mask):
                print("WARNING: lost %d" % (len(mask) - len(idx_set)) + \
                      " particles in snapshot masking!", file=sys.stderr)
            mask = np.array([(i in idx_set) for i in range(len(base))])
            return SubSnapshot(base, mask)

        elif isinstance(mask, tuple) and len(mask) == 1 \
                and isinstance(mask[0], np.ndarray) and mask[0].dtype.kind == 'i':
            return SubSnapshot(base, mask[0])

        else:
            raise KeyError('Mask of type %s not understood.' % type(mask).__name__)

    def write(self, filename, **kwargs):
        '''
        Write this (sub-)snapshot to the given file.

        For more information see `gadget.write` as this function just calls it.

        Args:
            filename (str):     The path to write the snapshot to.
            **kwargs:           Further (keyword) arguments are simply passed to
                                `gadget.write` (by default it is
                                `gad_units=self.root._gad_units`).
        '''
        return gadget.write(self, filename, gad_units=self.root._gad_units, **kwargs)

    def headers(self, idx=None):
        '''
        A list of the header(s) as dict's. For common features see `properties`.

        Args:
            idx (int):      If not None, do not return the entire list, but just
                            the header of the file of given number.
        '''
        if idx is None:
            return [reader.header.copy() for reader in self._root._file_handlers]
        else:
            return self._root._file_handlers[idx].header.copy()

    def families(self):
        '''Return the names of the particle families (at least partly) present in
        this snapshot.'''
        families = set()
        for name, ptypes in gadget.families.items():
            if sum(self._N_part[pt] for pt in ptypes):
                families.add(name)
        return list(families)

    def available_blocks(self):
        '''The names of the blocks that are available for all particle types of
        this snapshot.'''
        return [ name for name, ptypes
                in self._root._block_avail.items()
                if all([(Np==0 or s) for s,Np in zip(ptypes,self._N_part)]) ]

    def loadable_blocks(self):
        '''The names of all blocks of this snapshot.'''
        return list(self._root._load_name.keys())

    def deriveable_blocks(self):
        '''The names of all blocks of this snapshot.'''
        return list(self._root._derive_rule_deps.keys())

    def all_blocks(self):
        '''The names of all blocks of this snapshot.'''
        return list(self._root._block_avail.keys())

    def cosmic_time(self):
        '''The cosmic time (i.e. the current universe age).'''
        return self.cosmology.cosmic_time(self.redshift)

    def __repr__(self):
        return '<Snap %s; N=%s; z=%.3f>' % (
                self._descriptor,
                utils.nice_big_num_str(len(self)),
                self._root.redshift)

    def fill_derived_rules(self, rules=None, clear_old=False):
        '''
        Fill the derived rules for (the root of) this snapshot.

        Args:
            rules (dict):       A dictionary of the rules to add. The keys are the
                                names of the blocks (blocks that can be loaded
                                will not be added!). The entries shall be strings
                                with the rule.
                                Defaults to the list loaded from `derived.cfg`.
            clear_old (dict):   Whether to first clear any already existing rules.
        '''
        if not self is self.root:
            self.root.fill_derived_rules(rules, clear_old)
            return

        if rules is None:
            rules = derived._rules
        rules = rules.copy()

        if clear_old:
            for name in self._derive_rule_deps.keys():
                self._block_avail.pop(name,None)
            self._derive_rule_deps = {}

        # calculate the dependencies and particle types of the derived blocks
        # define 'dV' by 'vol_def_x' if not explicitly given
        if 'dV' not in rules:
            x = gadget.config.general['vol_def_x']
            if x != '<undefined>':
                rules['dV'] = '%s / kernel_weighted(gas,%s)' % (x,x)

        # remove derived blocks that can be loaded
        for name in list(rules.keys()):
            if name in self._load_name:
                del rules[name]

        # calculate the dependencies
        changed = True
        while changed:
            changed = False
            for name, rule in rules.items():
                if name in self._load_name:
                    continue    # this derived block can actually be loaded
                ptypes, deps = derived.ptypes_and_deps(rule, self)
                if name in self._block_avail:
                    if ptypes!=self._block_avail[name] \
                            or deps!=self._derive_rule_deps[name][1]:
                       changed = True
                else:
                   changed = True
                self._block_avail[name] = ptypes
                self._derive_rule_deps[name] = (rule, deps)

        # now remove those derived blocks, that depend on non-existent blocks
        not_available = set()
        removed = -1
        while removed != len(not_available):
            removed = len(not_available)
            for name, (rule, deps) in self._derive_rule_deps.items():
                if not (any(self._block_avail[name]) and (deps-not_available)):
                    not_available.add(name)
            for name in not_available:
                self._block_avail.pop(name,None)
                self._derive_rule_deps.pop(name,None)

        self._always_cache = set(self._derive_rule_deps.keys()) \
                            & derived.general['always_cache']

    def __dir__(self):
        # Add the families available such that they occure in tab completion in
        # iPython, for instance.
        # The other list added here seems to include excatly the properties.
        return [k for k in list(self.__dict__.keys())
                        if not k.startswith('_')] + \
                [k for k in list(self._root.__class__.__dict__.keys())
                        if not k.startswith('_')] + \
                self.families()

    def get_host_subsnap(self, block_name):
        '''
        Find the (sub-)snapshot, that stores a given block.

        The family sub-snapshot of the root that can use the block unsliced/-mask
        is the one to store it. Needless to say, if the block is available for all
        particle types, it is the root.

        Args:
            block_name (str):   The (attribute) name of the block.

        Returns:
            host (Snap):        The host (sub-)snapshot.
        '''
        root = self._root
        parts = root._N_part
        # It might happen, that a block is tagged as present for some particle
        # types that does not exist at all. Hence, the following:
        avail = root._block_avail[block_name]
        ptypes = [pt for pt in range(6) if (avail[pt] and parts[pt])]

        # block available for all families? -> root
        if not any( [parts[pt] for pt in range(6) if pt not in ptypes] ):
            return root

        # find biggest family sub-snapshot (that is an attribute of root) which has
        # the block
        for fam, fptypes in gadget.families.items():
            if fptypes == ptypes:
                return getattr(root, fam)

        return None

    def iter_blocks(self, present=True, block=True, name=False, forthis=True):
        '''
        Iterate all blocks, even those that are not for all particles.

        If present==False and block==True, this loads / derives all the not yet
        present (loaded/derived) blocks.

        Args:
            present (bool): If set, only iterate over the already loaded / derived
                            blocks.
            block (bool):   Whether to yield the (full) block itself.
            name (bool):    Whether to yield the block name.
            forthis (bool): Only include blocks that are available for this entire
                            (sub-)snapshot.

        Yields:
            block (UnitArr):    The full block, if only this was requested,
            name (str):         the name of the block, if only that was requested,
            OR
            (block, name):      both as a tuple, if both are requested.
            (And just nothing, if neither was requested.)
        '''
        done = set()
        if forthis:
            names = iter(self._blocks.keys()) if present else self.available_blocks()
        else:
            names = iter(self._root._block_avail.keys())
        for block_name in names:
            if block_name in done:
                continue
            host = self.get_host_subsnap(block_name)
            if host is None:
                continue
            if not present or block_name in host._blocks:
                if block and name:
                    yield host[block_name], block_name
                elif block:
                    yield host[block_name]
                elif name:
                    yield block_name

    def _get_block(self, name):
        '''
        This gets called, if a block is not yet set in self._blocks.

        Overwritten for _SubSnap, where care is taken for sub-snapshots that are
        not directly sliced/mased from the block's host. Here, for the root, it
        simply calls self._host_get_block (and asserts that the block is
        available).
        '''
        assert name in self.available_blocks()
        assert name not in self._blocks
        block = self._host_get_block(name)
        return block

    def __getattr__(self, name):
        # Create a family sub-snapshot and set it as attribute.
        if name in gadget.families:
            fam_snap = _FamilySubSnap(self, name)
            setattr(self, name, fam_snap)
            return fam_snap
        # Explicitly set attributes (like self.redshift and self.gas), member
        # functions, and properties are already handles before __getattr__ is
        # called, hence, we now have an attribute error:
        raise AttributeError('%r object has no ' % type(self).__name__ + \
                             'attribute %r' % name)

    def __getitem__(self, key):
        # strings are block names
        if isinstance(key, str):
            block = self._blocks.get(key,None)
            if block is not None:
                return block
            if key in self.available_blocks():
                # TODO: don't need to pass it though all the functions...
                return self._get_block(key)
            elif key in self._root._block_avail:
                raise KeyError('Block "%s" is not available for all ' % key +
                               'particle types of this (sub-)snapshot.')
            else:
                raise KeyError('(Sub-)Snapshot %r has no block "%s".' % (
                                    self, key))
        # postprone the import to here to speed up the access to blocks
        from .masks import SnapMask
        from ..analysis.halo import Halo
        if isinstance(key, (slice,np.ndarray,list,SnapMask,Halo,tuple)):
            # Handling of the index is fully done by the factory function.
            return self.SubSnap(key)
        else:
            raise KeyError(repr(key))

    def __setitem__(self, key, value):
        if key in self.available_blocks():
            old = self._blocks.get(key,None)
            if old is None:
                warnings.warn('Need to load/derive an available blocks before ' +
                              'overwriting it to some custom value!')
                old = self[key]
            from .sim_arr import SimArr
            if not isinstance(value, SimArr):
                value = SimArr(value, snap=self)
            if value.shape != old.shape:
                raise RuntimeError('Already existing blocks must not change ' +
                                   'their shape (here: "%s" with %r)!' % (
                                       key,old.shape))
            value.dependencies.update( old.dependencies )
            value.invalidate_dependencies()
            self._blocks[key] = value
        else:
            self._add_custom_block(value, key)

    def __delitem__(self, name):
        if name in self.available_blocks():
            try:
                host = self.get_host_subsnap(name)
                del host._blocks[name]
                if name not in self._root._load_name \
                        and name not in self._root._derive_rule_deps:
                    # delete custom blocks entirely
                    del self._root._block_avail[name]
            except KeyError:
                pass
        else:
            raise KeyError('(Sub-)Snapshot %r has no block "%s".' % (
                                    self, name))

    # iterate over all available loaded blocks (including custom and derived
    # blocks)
    def __iter__(self):
        for name, block in self.iter_blocks(present=True, block=True, name=True,
                                            forthis=True):
            yield name, block

    # whether a block of familiy is in the (sub-)snapshot (by name)
    def __contains__(self, el):
        return el in self.available_blocks() or el in self.families()

    def _host_load_block(self, name):
        '''
        Load a block and add it to its host.

        Note:
            For getting blocks -- also derived ones -- call _host_get_block!

        Args:
            name (str): The stripped name of the block (as it is in
                        self._root._load_name.keys()).

        Returns:
            block (SimArr):     The (entire) block.
        '''
        from .sim_arr import SimArr

        root = self._root
        if not root._file_handlers:
            raise RuntimeError('No file(s) to load from.')
        if not name in root._load_name:
            raise ValueError("There is no block '%s' to load!" % name)

        if environment.verbose >= environment.VERBOSE_NORMAL:
            print('load block %s%s...' % ('"%s" as '%root._load_name[name]
                    if environment.verbose >= environment.VERBOSE_TALKY else '',
                    name), end=' ')
            sys.stdout.flush()

        block_name = root._load_name[name]
        if len(root._file_handlers) == 1:
            block = root._file_handlers[0].read_block(block_name,
                                                      gad_units=root._gad_units)
        else:
            first = True
            for reader in root._file_handlers:
                # getting masses should always return something - even if there is
                # no such block, but the masses are given in the header only
                if not reader.has_block(block_name) and block_name!='mass':
                    continue
                read = reader.read_block(block_name, gad_units=root._gad_units)
                units = read.units
                if first:
                    blocks = [read[int(np.sum(reader.header['N_part'][:pt]))
                                    :np.sum(reader.header['N_part'][:pt+1])] \
                                for pt in range(6)]
                    first = False
                else:
                    for pt in range(6):
                        if not reader.header['N_part'][pt]:
                            continue
                        read_pt = read[int(np.sum(reader.header['N_part'][:pt]))
                                       :np.sum(reader.header['N_part'][:pt+1])]
                        blocks[pt] = np.concatenate((blocks[pt], read_pt),
                                                    axis=0).view(UnitArr)
                        blocks[pt].units = units
            assert not first
            block = np.concatenate(blocks, axis=0).view(UnitArr)
            block.units = units
        block = block.view(SimArr)
        block._snap = weakref.ref(self)
        if environment.verbose >= environment.VERBOSE_NORMAL:
            print('done.')
            sys.stdout.flush()

        if self.load_double_prec and block.dtype.kind == 'f' \
                and block.dtype != 'float64':
            if environment.verbose >= environment.VERBOSE_TALKY:
                print('convert to double precision...', end=' ')
                sys.stdout.flush()
            block = block.astype(np.float64)
            if environment.verbose >= environment.VERBOSE_TALKY:
                print('done.')
                sys.stdout.flush()

        if root._phys_units_requested:
            self._convert_block_to_physical_units(block, name)

        host = self.get_host_subsnap(name)
        host._blocks[name] = block

        for trans in root._trans_at_load:
            if name in trans._change:
                if environment.verbose >= environment.VERBOSE_NORMAL:
                    print('apply stored %s to block %s...' % (
                        trans.__class__.__name__, name), end=' ')
                    sys.stdout.flush()
                trans.apply_to_block(name,host)
                if environment.verbose >= environment.VERBOSE_NORMAL:
                    print('done.')
                    sys.stdout.flush()

        # if there are still dependent blocks, add them as dependencies and
        # invalidate them
        for derived_name, (rule, deps) in root._derive_rule_deps.items():
            if name in deps:
                host = self.get_host_subsnap(derived_name)
                if derived_name in host._blocks:
                    block.dependencies.add(derived_name)
        block.invalidate_dependencies()

        return block

    def _host_derive_block(self, name):
        '''
        Derive a block from the stored rules (and add it to its host).

        Note:
            For getting blocks call _host_get_block!

        Args:
            name (str):             The name of the derived block (as it is in
                                    self._root._derive_rule_deps).

        Returns:
            block (SimArr):     The (entire) block.
        '''
        root = self._root
        if not name in root._derive_rule_deps:
            raise ValueError("There is no block '%s' to derive!" % name)

        host = self.get_host_subsnap(name)
        rule, deps = root._derive_rule_deps[name]

        orig_caching_state = self._root._cache_derived
        for cache_name in self._root._always_cache:
            # in order to ensure that the automatic updating works (the dependencies
            # connect from the loading blocks to the always cached derived block),
            # temporarily turn on _cache_derived globally!
            if fnmatch.fnmatch(name, cache_name):
                self._root._cache_derived = True
                break

        # pre-load all needed (and not yet loaded) blocks in order not to mess up
        # the output (too much)
        dep_blocks = {}
        for dep in sorted(deps):
            if self._root._cache_derived:
                host[dep]
            else:
                # re-use stored blocks
                dep_blocks[dep] = host.get(dep, None)
                if dep_blocks[dep] is None:
                    # ... only calculate others, but do not store them
                    dep_blocks[dep] = host._host_derive_block(dep, False)

        if environment.verbose >= environment.VERBOSE_NORMAL:
            print('derive block %s%s...' % (name, ' := "%s"'%rule \
                    if environment.verbose >= environment.VERBOSE_TALKY else ''), end=' ')
            sys.stdout.flush()
        block = host.get(rule, namespace=None if self._root._cache_derived else dep_blocks)
        if self._root._cache_derived:
            for dep in deps:
                host[dep].dependencies.add(name)
            host._blocks[name] = block
        if environment.verbose >= environment.VERBOSE_NORMAL:
            print('done.')
            sys.stdout.flush()

        self._root._cache_derived = orig_caching_state

        return block

    def _add_custom_block(self, data, name):
        '''
        Add a new custom block to the snapshot.

        Args:
            data (array-like):  The new block (of appropiate size, i.e. length of
                                this (sub-)snapshot).
            name (str):         The name of the new block.

        Returns:
            block (SimArr):     The reference to the new block.

        Raises:
            RuntimeError:       If the length of the data does not fit the length
                                of the (sub-)snapshot; or if the latter is not the
                                correct "host" of the new block.
            KeyError:           If there already exists a block of that name.
        '''
        if len(data) == len(self):
            ptypes = [bool(N) for N in self.parts]
        else:
            raise RuntimeError('Data length does not fit the (sub-)snapshot!')

        if name in self._root._block_avail:
            raise NotImplementedError('Block name "%s" already exists ' % name +
                                      'for other parts of the snapshot!\n' +
                                      'Combining is not implemented.')

        self._root._block_avail[name] = list(ptypes)
        host = self.get_host_subsnap(name)
        if host is not self:
            print("name:", name, file=sys.stderr)
            print("for particle types:", self._root._block_avail[name], file=sys.stderr)
            print("self:", self, file=sys.stderr)
            print("host:", host, file=sys.stderr)
            del self._root._block_avail[name]
            raise RuntimeError('Can only set new blocks for entire family '
                               'sub-snapshots or the root.')

        from .sim_arr import SimArr
        host._blocks[name] = SimArr(data,snap=host)

    def _host_get_block(self, name):
        '''
        Load or derive a block of the given name and set it as attribute of its
        host.

        Args:
            name (str):         The stripped name of the block (as it is in
                                self._root._block_avail).

        Returns:
            block (SimArr):     The (entire) block.
        '''
        if name in self._root._load_name:
            return self._host_load_block(name)
        elif name in self._root._derive_rule_deps:
            return self._host_derive_block(name)
        else:
            raise ValueError("These is no block '%s' than can be " % name +
                             "loaded or derived!")

    def _convert_block_to_physical_units(self, block, name):
        '''
        Helper function to convert a block into its physical units, i.e. removing
        'a' and 'h_0' from the units.
        '''
        if block.units is None:
            return

        phys_units = block.units.free_of_factors(['a','h_0'])

        if phys_units != block.units:
            if environment.verbose >= environment.VERBOSE_TALKY or (
                    environment.verbose >= environment.VERBOSE_NORMAL
                    and name=='age'):
                print('convert block %s to physical units...' % name, end=' ')
                sys.stdout.flush()
            block.convert_to(phys_units, subs=self)
            if environment.verbose >= environment.VERBOSE_TALKY:
                print('done.')
                sys.stdout.flush()

    def to_physical_units(self, on_load=True):
        '''
        Convert all blocks and the boxsize to physical units.

        Convert all loaded blocks to units without scale factors and Hubble
        parameters and without numerical factors (e.g. 'a kpc / h_0' -> 'kpc' and
        '1e10 Msol / h_0' -> 'Msol').

        Note:
            There is no function to conveniently reverse this. You would have to
            unload (delete) all blocks, set phys_units_requested to False, and the
            reload the blocks (which is done automatically):
            --> for name, block in self:
            ...     del self[name]
            --> self.phys_units_requested = False

        Args:
            on_load (bool):     If True also convert newly loaded blocks
                                automatically.
        '''
        root = self._root

        # remember whether to convert to blocks or not
        root._phys_units_requested = bool(on_load)

        # convert all loaded blocks
        for block, name in self.iter_blocks(present=True, block=True, name=True,
                                            forthis=False):
            self._convert_block_to_physical_units(block, name)

        # convert the boxsize
        if environment.verbose >= environment.VERBOSE_TALKY:
            print('convert boxsize to physical units...', end=' ')
            sys.stdout.flush()
        root._boxsize.convert_to(
                root._boxsize.units.free_of_factors(['a','h_0']), subs=root)
        if environment.verbose >= environment.VERBOSE_TALKY:
            print('done.')
            sys.stdout.flush()

    def load_all_blocks(self):
        '''Load all blocks from file. (No block deriving, though.)'''
        for name in self._root._load_name:
            host = self.get_host_subsnap(name)
            host[name]

    def delete_blocks(self, derived=None, loaded=None):
        '''
        Delete all blocks.

        Args:
            derived, loaded (bools):
                        If both are None (the default), all blocks are deleted.
                        Otherwise only the blocks for which the argument is set to
                        True are deleted.
        '''
        if derived is None and loaded is None:
            derived = True
            loaded = True

        if derived:
            for name, (rule, deps) in self._root._derive_rule_deps.items():
                host = self.get_host_subsnap(name)
                if name in host._blocks:
                    del host[name]
        if loaded:
            for name in self._root._load_name:
                host = self.get_host_subsnap(name)
                if name in host._blocks:
                    del host[name]

    def get(self, expr, units=None, namespace=None):
        '''
        Evaluate the expression for a new block (e.g.: 'log10(Fe/mass)').

        The evaluation is done with a utils.Evaluator with numpy for math and the
        snapshot's blocks as well as some chosen objects (e.g. 'dist', 'Unit',
        'UnitArr', 'UnitQty', 'inter_bc_qty', 'solar', 'G') as namespace.

        Args:
            expr (str):         An python expression that knows about the blocks
                                in this snapshot.
            units (str, Unit):  If set, convert the array to this units before
                                returning.
            namespace (dict):   Additional namespace to use for calculating the
                                defined block.

        Returns:
            res (SimArr):       The result.
        '''
        from .sim_arr import SimArr
        from ..ssp import inter_bc_qty, lum_to_mag

        # prepare evaluator
        from numpy.core.umath_tests import inner1d
        from .. import analysis
        namespace = {} if namespace is None else namespace.copy()
        namespace.update( {'dist':dist, 'Unit':Unit, 'Units':Units,
                           'UnitArr':UnitArr, 'UnitQty':UnitQty,
                           'UnitScalar':UnitScalar, 'inner1d':inner1d,
                           'inter_bc_qty':inter_bc_qty, 'lum_to_mag':lum_to_mag,
                           'perm_inv':utils.perm_inv,
                           'solar':physics.solar, 'WMAP7':physics.WMAP7,
                           'Planck2013':physics.Planck2013,
                           'FLRWCosmo':physics.FLRWCosmo, 'a2z':physics.a2z,
                           'z2a':physics.z2a,
                           'kernel_weighted':analysis.kernel_weighted,
                           'len':len, 'module_dir':module_dir}
        )
        from . import derive_rules
        for n,obj in [(n,getattr(derive_rules,n)) for n in dir(derive_rules)]:
            if hasattr(obj, '__call__'):
                namespace[n] = obj
        for n,obj in [(n,getattr(physics,n)) for n in dir(physics)]:
            if isinstance(obj, UnitArr):
                namespace[n] = obj
        e = utils.Evaluator(namespace, my_math=np)
        # load properties etc. from this snapshot only if needed in the expression
        namespace = {'self':self}
        for name, el in utils.iter_idents_in_expr(expr, True):
            if name not in e.namespace and name not in namespace:
                try:
                    try:
                        namespace[name] = getattr(self,name)
                    except AttributeError:
                        namespace[name] = self[name]
                except KeyError:
                    if not isinstance(el, ast.Attribute):
                        raise ValueError("A unknown name ('%s') " % name +
                                         "appeared in expression for a new" +
                                         "block: '%s'" % expr)
        res = e.eval(expr, namespace)
        res = res.view(SimArr)
        res._snap = weakref.ref(self)
        if units is not None:
            res.convert_to(units)

        return res

    def IDs_unique(self):
        '''Check whether the IDs (of this (sub-)snapshot) are unique.'''
        return len(np.unique(self['ID'])) == len(self)


class SubSnapshot(Snapshot):
    '''
    A class for "sub-snapshots", that slice or mask an underlying (sub-)snapshot
    as a whole.

    The slice / mask is basically propagated to the arrays of the base. To
    instantiate this class, one should use the factory function SubSnap.

    Note:
        Sub-snapshots are convenient, but can become quite expensive in memory
        (and CPU time). If you encounter problems, consider taking the needed
        sub-blocks from it only and deleting the sub-snapshots -- or masking the
        blocks directly. Making copies can eliminate the mask which might occupy a
        lot of memory.

    Args:
        base (Snap):    The base (sub-)snapshot.
        mask (slice, np.ndarray[bool]):
                        A slice or a mask in form of a boolean np.ndarray.
        N_part (list):  The number of particles per type to take. Can be faster
                        than calculating them by the standard functions.
                        ATTENTION: you are responsible for this being correct!

    Doctests:
        >>> s = Snapshot(module_dir+'snaps/snap_M1196_4x_470', physical=False)
        >>> if any(s.gas.parts[1:]): print(s.gas.parts)
        >>> if any(s.stars.parts[:4]) or any(s.stars.parts[5:]): print(s.gas.parts)
        >>> s[123:len(s):100]
        <Snap "snap_M1196_4x_470"[123:2079055:100]; N=20,790; z=0.000>
        >>> s[123:len(s):100].parts
        [9216, 10015, 568, 193, 798, 0]
        >>> s['pos'] -= [34623, 35522, 33181]
        load block pos... done.
        >>> mask = abs(s['pos']).max(axis=-1) < 100
        >>> s[mask]
        <Snap "snap_M1196_4x_470":masked; N=355,218; z=0.000>
        >>> s[mask][::100].parts
        [431, 2459, 0, 0, 663, 0]
        >>> s[mask][::100].gas
        <Snap "snap_M1196_4x_470":masked[::100]:gas; N=431; z=0.000>
        >>> slim = s[::100]
        >>> slim.parts
        [9218, 10014, 568, 193, 798, 0]
        >>> np.sum(slim.gas['mass'] >= 0.0002)
        load block mass... done.
        UnitArr(1)
        >>> slim_mask = slim['mass'] < 0.0002   # exludes one gas particle...
        >>> slim[slim_mask].parts
        [9217, 0, 0, 0, 798, 0]
        >>> slim[slim_mask][100:9000:3]['rho']
        load block rho... done.
        SimArr([3.1859734e-10, 8.6869852e-11, 2.5398775e-10, ..., 9.3606489e-10,
                1.1147246e-09, 4.2680767e-10],
               dtype=float32, units="1e+10 Msol ckpc**-3 h_0**2", snap="snap_M1196_4x_470":gas)
        >>> len(slim[slim_mask][100:9000:3]['rho'])
        2967
        >>> mask = np.zeros(len(s), dtype=bool)
        >>> for pt in gadget.families['baryons']:
        ...     mask[sum(s.parts[:pt]):sum(s.parts[:pt+1])] = True

        s[mask], however, is not host (!)
        >>> assert np.all( s[mask]['elements'] == s.baryons['elements'] )
        load block Z... done.
        derive block elements... done.
        >>> len(slim[slim_mask]['elements'])
        10015
        >>> assert s[2:-123:10].stars['form_time'].shape[0] == len(s[2:-123:10].stars)
        load block form_time... done.
        >>> s[2:-123:10].stars['form_time']
        SimArr([0.7739953 , 0.950091  , 0.81303227, ..., 0.2623567 , 0.40703428,
                0.28800005],
               dtype=float32, units="a_form", snap="snap_M1196_4x_470":stars)
        >>> assert slim[slim_mask].stars['form_time'].shape[0] == len(slim[slim_mask].stars)
        >>> slim[slim_mask].stars['form_time'][::100]
        SimArr([0.9974333 , 0.30829304, 0.6479812 , 0.6046823 , 0.18474203,
                0.5375404 , 0.18224117, 0.26500258],
               dtype=float32, units="a_form", snap="snap_M1196_4x_470":stars)

        # index list (+family) tests (see comments in SubSnap)
        >>> r = np.sqrt(np.sum(s['pos']**2, axis=1))
        >>> assert np.all( s[r<123]['pos'] == s[np.where(r<123, True, False)]['pos'] )
        >>> assert np.all( s[s['pos'][:,0]>0]['pos']
        ...             == s[np.where(s['pos'][:,0]>0, True, False)]['pos'] )
        >>> assert np.all( s.stars[s.stars['inim']>s.stars['mass'].mean()]['pos'] ==
        ...         s.stars[np.where(s.stars['inim']>s.stars['mass'].mean(), True, False)]['pos'] )
        load block inim... done.
    '''
    def _calc_N_part_from_slice(self, N_part_base, _slice):
        # work with positive stepping (and without None's):
        _slice = utils.sane_slice(_slice, int(sum(N_part_base)))
        start, stop, step = _slice.start, _slice.stop, _slice.step

        N_part = [0]*6
        cur_pos = start
        N_cum_base = np.cumsum(N_part_base)
        pt = 0
        # goto first particle type that is in the sub-snapshot
        while N_cum_base[pt] <= cur_pos:
            pt += 1
        # count particles of the different types
        while pt < 6 and cur_pos < stop:
            remain = int( min(stop, N_cum_base[pt]) - cur_pos )
            N_part[pt] = remain // step
            if remain % step:
                N_part[pt] += 1
            cur_pos += N_part[pt] * step
            pt += 1

        return N_part

    def _calc_N_part_from_mask(self, N_part_base, mask):
        N_cum_base = np.cumsum(N_part_base)
        return [np.sum(mask[s:e]) for s,e in zip([0]+list(N_cum_base[:-1]),
                                                 N_cum_base)]

    def __init__(self, base, mask, N_part=None):
        #Snapshot.__init__(self, gad_units=None, physical=False, cosmological=None)
        self._base      = base
        self._root      = base._root
        self._blocks    = {}

        if isinstance(mask, slice):
            if mask.step == 0:
                raise ValueError('Slice step cannot be 0!')

            self._mask = slice(mask.start, mask.stop, mask.step) # make a copy
            if N_part is None:
                self._N_part = self._calc_N_part_from_slice(base._N_part, mask)
            else:
                self._N_part = N_part[:]

            self._descriptor = base._descriptor + '['
            if mask.start is not None:
                self._descriptor += str(mask.start)
            self._descriptor += ':'
            if mask.stop is not None:
                self._descriptor += str(mask.stop)
            if mask.step is not None:
                self._descriptor += ':' + str(mask.step)
            self._descriptor += ']'
        elif isinstance(mask, np.ndarray) and mask.dtype == bool:
            mask = mask.view(np.ndarray)
            if len(mask) != len(base):
                raise ValueError('Mask has to have the same length as the ' + \
                                 'base snapshot!')

            self._mask = mask.copy()
            if N_part is None:
                self._N_part = self._calc_N_part_from_mask(base._N_part, mask)
            else:
                self._N_part = N_part[:]

            self._descriptor = base._descriptor + ':masked'
        else:
            raise TypeError('Need either a slice or a boolean np.ndarray ' + \
                            'for sub-snapshots!')

    def _restrict_mask_to_ptypes(self, ptypes):
        '''
        Get the sub-snapshot's mask restricted to the given particle types.

        Args:
            ptypes (list):  A (sorted) list of the particle types to restrict to.
        '''
        base = self._base

        if isinstance(self._mask, slice):
            # work with positive stepping (and without None's):
            _slice = utils.sane_slice(self._mask, len(base))
            start, stop, step = _slice.start, _slice.stop, _slice.step
            offset = sum( base._N_part[:ptypes[0]] )
            if utils.is_consecutive(ptypes):
                if start < sum(base._N_part[:ptypes[0]]):
                    diff = sum(base._N_part[:ptypes[0]]) - start
                    start += diff // step * step
                    if diff % step:
                        start += step
                stop = min(stop, sum(base._N_part[:ptypes[-1]+1]))
                return slice(start-offset, stop-offset, step)
            else:
                # slicing is not sufficient, need masking
                fullmask = np.zeros(len(base), dtype=bool)
                fullmask[start:stop:step] = True
                pmasks = []
                for pt in ptypes:
                    start = int(sum(base._N_part[:pt]))
                    end   = start + base._N_part[pt]
                    pmasks.append( fullmask[start:end] )
                mask = np.concatenate(pmasks)
                return mask
        else:
            assert isinstance(self._mask, np.ndarray) and self._mask.dtype==bool
            if utils.is_consecutive(ptypes):
                start = sum(base._N_part[:ptypes[0]])
                end   = sum(base._N_part[:ptypes[-1]+1])
                return self._mask[start:end]
            else:
                pmasks = []
                for pt in ptypes:
                    start = int(sum(base._N_part[:pt]))
                    end   = start + base._N_part[pt]
                    pmasks.append( self._mask[start:end] )
                mask = np.concatenate(pmasks)
                return mask

    def _get_block(self, name):
        '''
        This gets called, if a block is not yet set in self._blocks.

        Here for a _SubSnap, one has to handle the case where the block is not
        available for the base. It has to be stored appropiately then.
        '''
        if name in self._base.available_blocks():
            # do not store the masked snapshot, but only the base one
            return self._base[name][self._mask]
        else:
            # Due to calling by __getitem__, the following should always be
            # fulfilled:
            assert name in self.available_blocks()
            assert name not in self._blocks
            host = self.get_host_subsnap(name)
            if host is self:
                # no slicing/masking needed!
                # (caching of blocks is done within this function)
                block = host._host_get_block(name)
                return block
            else:
                # delegate the loading to the host
                block = host[name]
                #################################################################
                # The general situation here:
                #
                #   root -- sub[1] -- ... -- sub[N-1] -- sub[N] (==self)
                #       \
                #       host
                #
                # where sub[1] != host and sub[N-1] does not have the requested
                # block.
                # Now restrict all the masks of sub[1] though sub[N] and self to
                # the particle types of the host and apply the masks to the block.
                #################################################################
                ptypes = [pt for pt in range(6) if host._N_part[pt]]
                masks = []
                sub = self
                subs = []
                while sub != sub._root:
                    masks.append( sub._restrict_mask_to_ptypes(ptypes) )
                    subs.append( sub )
                    sub = sub._base
                for mask in reversed(masks):
                    block = block[mask]
                # do not store the masked snapshot, but only the base one
                return block


def _FamilySubSnap(base, fam):
    '''
    A factory function for creating a sub-snapshot of all the particles of the
    given family of the base snapshot.

    Sub-snapshots for families should always be instantiated with this function or
    by accessing attributes. The latter caches the sub-snapshot.

    Args:
        base (Snap):        The snapshot to create the family sub-snapshot from.
        fam (str):          The name of the family (as in `gadget.families`) to
                            restrict to.

    Returns:
        fam_subsnap (Snap): The sub-snapshot of `base` that is restricted to the
                            specified family.
    '''
    if fam not in gadget.families:
        raise ValueError('Unknown familiy "%s"!' % fam)

    # family sub-snapshot
    family_s = base.SubSnap(gadget.families[fam])
    family_s._descriptor = base._descriptor + ':' + fam
    return family_s


def Snap(filename, physical=False, load_double_prec=False, cosmological=None,
             gad_units=None, unclear_blocks=None, H_neutral_only=None):
    from warnings import warn
    warn("\nfunction Snap(...) has been renamed to Snapshot class\nplease use Snapshot(..) to create a snapshot object", DeprecationWarning, stacklevel=2)
    s = Snapshot(filename, physical, load_double_prec, cosmological, gad_units, unclear_blocks, H_neutral_only)
    return s


'''
Classes for snapshot transformations.

Examples:
    >>> from ..environment import module_dir
    >>> from ..snapshot import Snapshot
    >>> rot = Rotation([[0,1,0],[1,0,0],[0,0,1]])
    Traceback (most recent call last):
    ...
    ValueError: Rotation matrix needs to fullfil det(R) == +1 (proper Rotation)!
    >>> rot = Rotation([[0,1,0],[0,0,1],[1,0,0]])
    >>> s = Snapshot(module_dir+'snaps/snap_M1196_4x_470', physical=False)
    >>> s['pos'][:3]
    load block pos... done.
    SimArr([[34613.516, 35521.816, 33178.605],
            [34613.297, 35521.766, 33178.316],
            [34613.26 , 35521.883, 33178.48 ]],
           dtype=float32, units="ckpc h_0**-1", snap="snap_M1196_4x_470")
    >>> rot.apply(s)
    apply Rotation to "pos" of "snap_M1196_4x_470"... done.
    >>> s['pos'][:3]
    SimArr([[35521.816, 33178.605, 34613.516],
            [35521.766, 33178.316, 34613.297],
            [35521.883, 33178.48 , 34613.26 ]],
           dtype=float32, units="ckpc h_0**-1", snap="snap_M1196_4x_470")
    >>> s['vel'][:3]
    load block vel... done.
    apply stored Rotation to block vel... done.
    SimArr([[-119.96177, -215.2135 , -218.84106],
            [-100.0279 , -203.8982 , -218.84097],
            [-107.25678, -204.94604, -213.70818]],
           dtype=float32, units="a**1/2 km s**-1", snap="snap_M1196_4x_470")
    >>> Translation(UnitArr([10,-20,30],'Mpc')).apply(s)
    apply Translation to "pos" of "snap_M1196_4x_470"... done.
    >>> s['pos'][:3]
    SimArr([[42721.816, 18778.605, 56213.516],
            [42721.766, 18778.316, 56213.297],
            [42721.883, 18778.48 , 56213.26 ]],
           dtype=float32, units="ckpc h_0**-1", snap="snap_M1196_4x_470")
    >>> Translation(UnitArr([100,200,300])).apply(s)
    apply Translation to "pos" of "snap_M1196_4x_470"... done.
    >>> s['angmom']
    load block mass... done.
    derive block momentum... done.
    derive block angmom... done.
    SimArr([[ 453.28464 ,  146.67963 , -392.72455 ],
            [ 627.0708  ,  316.3762  , -581.3953  ],
            [ 483.95087 ,  198.68784 , -433.42886 ],
            ...,
            [  59.79834 , -180.23755 ,   15.165218],
            [ 147.31934 ,  -25.509079, -103.11534 ],
            [  60.93218 ,  114.03373 ,  -84.469315]],
           dtype=float32, units="1e+10 Msol a**1/2 ckpc h_0**-1 h_0**-1 km s**-1", snap="snap_M1196_4x_470")
    >>> rot.apply(s)
    apply Rotation to "pos" of "snap_M1196_4x_470"... done.
    apply Rotation to "vel" of "snap_M1196_4x_470"... done.
    >>> s['angmom']
    derive block momentum... done.
    derive block angmom... done.
    SimArr([[ 146.67963 , -392.72455 ,  453.28464 ],
            [ 316.3762  , -581.3953  ,  627.0708  ],
            [ 198.68784 , -433.42886 ,  483.95087 ],
            ...,
            [-180.23755 ,   15.165218,   59.79834 ],
            [ -25.509079, -103.11534 ,  147.31934 ],
            [ 114.03373 ,  -84.469315,   60.93218 ]],
           dtype=float32, units="1e+10 Msol a**1/2 ckpc h_0**-1 h_0**-1 km s**-1", snap="snap_M1196_4x_470")
    >>> ca, sa = np.cos(12), np.sin(12)
    >>> Rotation([[ca,sa,0],[-sa,ca,0],[0,0,1]]).apply(s)
    apply Rotation to "pos" of "snap_M1196_4x_470"... done.
    apply Rotation to "vel" of "snap_M1196_4x_470"... done.
    >>> for T in s._trans_at_load:  # for efficiency, translations of same type
    ...     print(T)                 # are combined...
    Rotation([[ 0.000, 1.000, 0.000],[ 0.000, 0.000, 1.000],[ 1.000, 0.000, 0.000]])
    Translation([10.14,-19.72,30.42] [Mpc])
    Rotation([[ 0.000, 0.844,-0.537],[ 0.000, 0.537, 0.844],[ 1.000, 0.000, 0.000]])
    >>> old = s['pos'][:3]
    >>> del s['pos']
    >>> assert np.max(np.abs((s['pos'][:3] - old) / old)) < 1e-6
    load block pos... done.
    apply stored Rotation to block pos... done.
    apply stored Translation to block pos... done.
    apply stored Rotation to block pos... done.
    >>> del s; s = Snapshot(module_dir+'snaps/snap_M1196_4x_470', physical=False)
    >>> print(s['pos'][0]); print(s['pos'][-1])
    load block pos... done.
    [34613.516 35521.816 33178.605] [ckpc h_0**-1]
    [34607.652 35537.62  33167.832] [ckpc h_0**-1]
    >>> rot.apply(s.gas,remember=False)
    apply Rotation to "pos" of "snap_M1196_4x_470":gas... done.
    >>> print(s['pos'][0]); print(s['pos'][-1])
    [35521.816 33178.605 34613.516] [ckpc h_0**-1]
    [34607.652 35537.62  33167.832] [ckpc h_0**-1]
    >>> del s['pos']

    >>> np.linalg.norm(s['pos'][0])
    load block pos... done.
    59671.727
    >>> rot_to_z([1,0,0]).apply(s.gas, total=True)
    apply Rotation to "pos" of "snap_M1196_4x_470"... done.
    >>> np.linalg.norm(s['pos'][0])
    59671.727
    >>> s['vel'][-3:]
    load block vel... done.
    apply stored Rotation to block vel... done.
    SimArr([[ 14.277103 , -56.81834  ,  54.359398 ],
            [ 37.947582 ,  84.524315 , 130.14624  ],
            [ 37.378033 ,  64.147415 ,  -4.1874957]],
           dtype=float32, units="a**1/2 km s**-1", snap="snap_M1196_4x_470")'''
__all__ = ['Transformation', 'Translation', 'Rotation', 'rot_from_axis_angle',
           'rot_to_z']

import sys
import numpy as np
from ..units import *
from .. import environment

class Transformation(object):
    '''
    The base-class for transformations.

    Args:
        change (dict):  A dictionary from the block names to the functions to call
                        when the transformation is applied to them in self.apply.
                        They have to have the call signature f(name,snap).
    '''

    def __init__(self, change, pre=None, post=None):
        self._change = change.copy()
        self._pre  = [] if pre  is None else pre.copy()
        self._post = [] if post is None else prost.copy()

    def __copy__(self):
        cp = Transformation.__new__(Transformation)
        cp._change = self._change.copy()
        cp._pre  = self._pre.copy()
        cp._post = self._prost.copy()
        return cp

    @property
    def changes(self):
        '''The names of the blocks that are changed, if applied to a snapshot.'''
        return list(self._change.keys())

    def _apply_to_block(self, block, snap=None):
        raise NotImplementedError('This is just an interface / a virtual base '
                                  'class!')

    def apply_to_block(self, block, snap=None):
        '''
        Apply the transformation to a single block.

        Args:
            block (np.ndarray, str):    The block to apply the transformation to.
                                        If snap==None, this has to be an array,
                                        which then is not changed in-place; else
                                        if snap is a snapshot, it has to be a
                                        string an the block is transformed
                                        in-place (if possible) in the snapshot.
            snap (Snap):                The snapshot the block belongs to (c.f.
                                        above).

        Returns:
            transformed (array-like):   The transformed block (also if changed
                                        in-place).
        '''
        if isinstance(block,str) and block in self._change:
            return self._change[block](block,snap)
        else:
            return self._apply_to_block(block,snap)

    def apply(self, snap, total=False, remember=True, fams=True, exclude=None):
        '''
        Apply the transformation to all available and already loaded/derived
        blocks in the snapshot.

        Args:
            snap (Snap):    The snapshot to which the transformation shall be
                            applied. It can be a sub-snapshot, then the trans-
                            formation is only applied to that.
            total (bool):   Whether to apply the transformation to the entire
                            snapshot or just the passed sub-snapshot. A total=True
                            overwrites the argument snap to snap.root.
            remember (bool):
                            If set to True and the transformation is applied to
                            the root snapshot (which is always true if 'total' is
                            set), the snapshot remembers the transformation and
                            also applies it to newly loaded blocks.
            fams (bool):    Whether to include the blocks that are only available
                            for some of the particle types.
            exclude (set, list, tuple):
                            Some object that supports the operator 'in'. Block
                            names that are in exclude are skipped.
        '''
        if total and snap is not snap.root:
            snap = snap.root
        if exclude is None:
            exclude = set()

        done = set()
        # do blocks specified in 'self._pre' first and those in 'self._post' in
        # the end
        todo = self._pre \
             + list(set(self._change)-set(self._pre)-set(self._post)) \
             + self._post
        todo.sort()
        for name in todo:
            # only apply to available and not excluded blocks
            if name not in snap.available_blocks() or name in exclude:
                continue
            # ... and do not load blocks here
            if name not in snap._blocks:
                # might not yet be sliced, but loaded into its host
                if name not in snap.get_host_subsnap(name)._blocks:
                    continue
            if environment.verbose >= environment.VERBOSE_NORMAL:
                print('apply %s to "%s" of %s...' % (self.__class__.__name__,
                                                     name, snap.descriptor), end=' ')
                sys.stdout.flush()
            self._change[name](name,snap)
            if environment.verbose >= environment.VERBOSE_NORMAL:
                print('done.')
                sys.stdout.flush()
            done.add(name)

        if fams:
            for fam in [getattr(snap,fn) for fn in snap.families()]:
                # fams=True would result in infinite recursion
                # exclude=done prevents multiple applications on (partial) blocks
                # total=True would overwrite snap=fam
                # remember=True would not be effectual, since snap!=root
                self.apply(fam, total=False, remember=False, fams=False,
                           exclude=done)

        if remember and snap is snap.root:
            if snap._trans_at_load and type(snap._trans_at_load[-1])==type(self):
                last = snap._trans_at_load[-1]
                if isinstance(self, Translation):
                    trans = self._trans.copy()
                    if trans.units is None:
                        trans.units = snap.boxsize.units
                    last._trans += trans.in_units_of(last._trans.units,subs=snap)
                elif isinstance(self, Rotation):
                    last._R = self._R * last._R
                else:
                    raise NotImplementedError('Encountered an unknown '
                                              'transformation: %s' %
                                              self.__class__.__name__)
            else:
                if isinstance(self, Translation) and self._trans.units is None:
                    self._trans.units = snap.boxsize.units
                snap._trans_at_load.append( self.copy() )

class Translation(Transformation):
    '''
    A translation (in space).

    Args:
        trans (UnitQty):    The translation vector.

    Doctests:
        >>> T = Translation(UnitArr([1.0,-2.0,1.23],'kpc'))
        >>> T.trans
        UnitArr([ 1.  , -2.  ,  1.23], units="kpc")
        >>> Tcp = T.copy()
        >>> assert Tcp is not T
        >>> assert Tcp._trans is not T._trans
        >>> assert np.all(Tcp.trans == T.trans)
    '''

    def __init__(self, trans):
        # velocities are peculiar, hence already correct
        super(Translation,self).__init__( change={
                    'pos':  self._apply_to_block,
                })
        self.trans = trans

    def __copy__(self):
        return Translation(self.trans)

    def copy(self):
        return self.copy()

    @property
    def trans(self):
        return self._trans.copy()

    @trans.setter
    def trans(self, value):
        self._trans = UnitQty(value).astype(float).copy()
        if self._trans.shape != (3,):
            raise ValueError('The translation vector needs to be a 3-vector!')

    def copy(self):
        '''Copy the translation.'''
        return Translation(self._trans)

    def inverse(self):
        '''Calculate the inverse translation.'''
        return Translation(-self._trans)

    def __str__(self):
        unit_str = ' %s'%self._trans.units if self._trans.units else ''
        return 'Translation([%.4g,%.4g,%.4g]%s)'%(tuple(self._trans)+(unit_str,))

    def _apply_to_block(self, block, snap=None):
        if snap is not None:
            block = snap[block]
        if not len(block.shape)==2 and block.shape[1]==self._trans.shape[0]:
            raise ValueError('The block has to have shape (?,%d)!' %
                    self._trans.shape[0])

        if snap is None:
            block = block + self._trans
        else:
            block += self._trans.in_units_of(block.units,subs=snap)
        return block

class Rotation(Transformation):
    '''
    A spatial rotation.

    Args:
        rotmat (array-like):    The rotation matrix (has to fullfil R.T*R==1).
        test_proper (bool):     Test for a proper rotation (i.e. det(R) = +1) on
                                initialization and whenever the rotation matrix is
                                changed.

    Doctests:
        >>> ca, sa = np.cos(1.2), np.sin(1.2)
        >>> R = Rotation([[ca,sa,0],[-sa,ca,0],[0,0,1]])
        >>> R.rotmat
        matrix([[ 0.36235775,  0.93203909,  0.        ],
                [-0.93203909,  0.36235775,  0.        ],
                [ 0.        ,  0.        ,  1.        ]])
        >>> Rcp = R.copy()
        >>> assert Rcp is not R
        >>> assert Rcp._R is not R._R
        >>> assert np.all(Rcp.rotmat == R.rotmat)
        >>> assert Rcp.test_proper == R.test_proper
    '''

    def __init__(self, rotmat, test_proper=True):
        super(Rotation,self).__init__( change={
                    'pos':  self._apply_to_block,
                    'vel':  self._apply_to_block,
                    'acce': self._apply_to_block,
                } )
        self.test_proper = test_proper
        self.rotmat = rotmat

    def __copy__(self):
        return Rotation(self.rotmat, self.test_proper)

    def copy(self):
        return self.copy()

    @property
    def rotmat(self):
        return self._R.copy()

    @rotmat.setter
    def rotmat(self, value):
        self._R = np.matrix(value, dtype=float).copy()
        if len(self._R.shape)!=2 or self._R.shape!=(3,3):
            raise ValueError('Rotation matrix needs to be a 3x3-matrix!')
        if np.sum(np.abs( (self._R * self._R.T) - np.eye(3) )) > 1e-6:
            raise ValueError('Rotation matrix needs to fullfil R^T == R^-1!')
        if self._test_proper and not self.is_proper():
            raise ValueError('Rotation matrix needs to fullfil det(R) == +1 '
                             '(proper Rotation)!')

    @property
    def test_proper(self):
        return self._test_proper

    @test_proper.setter
    def test_proper(self, value):
        self._test_proper = bool(value)

    def is_proper(self):
        '''Test if the rotation is a proper rotation, i.e. it perserves
        chirality.'''
        return np.linalg.det(self._R) > 0.0

    def axis(self):
        '''Calculate the axis around which the rotation is performed.'''
        vals, vecs = np.linalg.eig(self._R)
        i = np.where(np.abs(vals-1) < 1e-9)[0][0]
        return vecs[:,i].real.getA1()

    def angle(self):
        '''Calculate the angle of the rotation.'''
        return np.arccos((np.trace(self._R) - 1.0) / 2.0)

    def copy(self):
        '''Copy the rotation.'''
        return Rotation(self._R)

    def inverse(self):
        '''Calculate the inverse rotation.'''
        return Rotation(self._R.T)  # remember: R^T == R^-1

    def __str__(self):
        mat = ''
        for i in range(3):
            mat += ',[%6.3f,%6.3f,%6.3f]' % (self._R[i,0],self._R[i,1],self._R[i,2])
        mat = '[%s]' % mat.strip(',')
        return 'Rotation(%s)' % mat

    def _apply_to_block(self, block, snap=None):
        if snap is not None:
            name, block = block, snap[block]
        if not len(block.shape)==2 and block.shape[1]==self._R.shape[0]:
            raise ValueError('The block has to have shape (?,%d)!' %
                    self._R.shape[0])

        transformed = (self._R * block.T).T.view(type(block))

        if isinstance(block, UnitArr):
            transformed.units = block.units
            transformed = transformed.view(UnitArr) # downgrade if SimArr

        if snap is not None:
            block[:] = transformed
        return transformed

def rot_from_axis_angle(u, angle):
    '''
    Create a Rotation of an angle around a axis.

    Args:
        u (array-like):         The axis to rotate around (does not have to be
                                normalized).
        angle (UnitScalar):     The angle in radians around the axis u (right-hand
                                rotation). (Default units: rad)

    Returns:
        rot (Rotation):         The corresponding rotation (not only the matrix!).

    Example:
        >>> print(rot_from_axis_angle([1,0,0], np.pi/2.0))
        Rotation([[ 1.000, 0.000, 0.000],[ 0.000, 0.000,-1.000],[ 0.000, 1.000, 0.000]])
        >>> print(rot_from_axis_angle([0,0,1], np.pi/4.0))
        Rotation([[ 0.707,-0.707, 0.000],[ 0.707, 0.707, 0.000],[ 0.000, 0.000, 1.000]])
        >>> print(rot_from_axis_angle([0,0,1], '12 degree'))
        Rotation([[ 0.978,-0.208, 0.000],[ 0.208, 0.978, 0.000],[ 0.000, 0.000, 1.000]])
        >>> R = rot_from_axis_angle([1,2,3], 1.234)
        >>> np.sqrt(1**2+2**2+3**2)*R.axis(), R.angle()
        (array([1., 2., 3.]), 1.234)
    '''
    if isinstance(angle, (str,UnitArr)):
        angle = UnitScalar(angle, 'rad')
    angle = float(angle)
    u = np.array(u, dtype=float)
    if u.shape != (3,):
        raise ValueError('The axis has to be a 3-vector.')
    norm = np.linalg.norm(u)
    if norm == 0:
        raise ValueError('The axis may not have zero norm!')
    u /= norm

    cos_a, sin_a = np.cos(angle), np.sin(angle)
    uxu = np.tensordot(u,u,0)
    u_x = np.array([[0,-u[2],u[1]],[u[2],0,-u[0]],[-u[1],u[0],0]])

    return Rotation(cos_a*np.eye(3)+sin_a*u_x+(1.0-cos_a)*uxu)

def rot_to_z(z):
    '''
    Calculate a rotation such that the new z-axis is the defined one.

    Args:
        z (array-like): The vector defining the new z-axis. (Does not have to be
                        normalized.)

    Returns:
        rot (Rotation): The corresponding rotation.

    Examples:
        >>> assert np.all(rot_to_z([0,0,1]).rotmat == np.eye(3))
        >>> z = rot_to_z([1,2,3]).apply_to_block(np.array([[1,2,3]]))
        >>> assert abs(z[0,0])<1e-9 and abs(z[0,1])<1e-9
        >>> print(rot_to_z([0,1,1]))
        Rotation([[ 1.000, 0.000, 0.000],[ 0.000, 0.707,-0.707],[ 0.000, 0.707, 0.707]])
        >>> print(rot_to_z([0.5,-1,0]))
        Rotation([[ 0.800, 0.400,-0.447],[ 0.400, 0.200, 0.894],[ 0.447,-0.894, 0.000]])
    '''
    z = np.array(z,dtype=float)
    if z.shape != (3,):
        raise ValueError('The new z axis has to be defined by a 3-vector.')
    norm = np.linalg.norm(z)
    if norm == 0:
        raise ValueError('The axis may not have zero norm!')
    z /= norm

    # get shortest rotation
    u = np.cross(z,[0,0,1])
    angle = np.arccos(z[2])

    if np.linalg.norm(u) < 1e-10:
        return Rotation(np.eye(3))

    return rot_from_axis_angle(u,angle)




'''
Module for low-level access to Gadget files of format 1, 2, and 3 (HDF5).

Writing snapshots is not yet fully supported and HDF5 files cannot be written at
all, yet.

Example:
    This tests format 2 with info block only.
    TODO: add other doctests!? Some are implicitly done in snapshot module.
    >>> from ..environment import module_dir
    >>> orig_file = module_dir+'snaps/snap_M1196_4x_470'
    >>> dest_file = module_dir+'test.gad'
    >>> blocks = {}
    >>> with open(orig_file, 'rb') as gfile:
    ...     gformat, endianness = get_format_and_endianness(gfile)
    ...     header = read_header(gfile, gformat, endianness)
    ...     info = get_block_info(gfile, gformat, endianness, header, 'exception')
    ...     for block in info.values():
    ...         N = sum(header['N_part'][i] for i in range(6) if block.ptypes[i])
    ...         gfile.seek(block.start_pos)
    ...         data = np.fromfile(gfile,
    ...                             dtype=endianness+block.type_descr,
    ...                             count=N*block.dimension)
    ...         if block.dimension > 1:
    ...             data = data.reshape( (N,block.dimension) )
    ...         blocks[block.name] = data
    300
    24948984
    49897668
    58213912
    66530156
    70217012
    73903868
    77590724
    81277580
    84964436
    88651292
    88970372
    89289452
    89608532
    137679212
    145995456
    >>> import os
    >>> assert not os.path.exists(dest_file)
    >>> info = sorted(info.values(), key=lambda x: x.start_pos)
    >>> with open(dest_file, 'wb') as gfile:
    ...     write_header(gfile, header, gformat, endianness)
    ...     for block in info:
    ...         write_block(gfile, block.name, blocks[block.name], gformat,
    ...                     endianness)
    ...     write_info(gfile, info, gformat, endianness)
    >>> import filecmp
    >>> assert filecmp.cmp(orig_file, dest_file)
    >>> os.remove(dest_file)
'''
__all__ = ['get_format_and_endianness', 'read_header', 'write_header',
           'BlockInfo', 'write_info', 'get_block_info', 'write_block']

import sys
from ..environment import secure_get_h5py
h5py = secure_get_h5py()
import struct
from . import config
import numpy as np
from itertools import combinations

SEEK_SET = 0
SEEK_CUR = 1
SEEK_END = 2

def get_format_and_endianness(gfile):
    '''
    Get format and endianness of the already opened gadget file.

    Args:
        gfile (file):       The already in binary read mode opened Gadget file or
                            a HDF5 file (which then is detected as format 3).

    Returns:
        gformat (int):      The format; either 1, 2, or 3.
        endianness (str):   Either '=' for native endianness or '<' / '>' for
                            non-native little / big endianness. None in case of a
                            HDF5 file.
    '''
    if isinstance(gfile, h5py.File):
        return 3, None

    gfile.seek(0)
    size, = struct.unpack('=i', gfile.read(4))
    if size == 256:
        gformat    = 1
        endianness = '='
    elif size == 8:
        gformat    = 2
        endianness = '='
        assert struct.unpack('=4s', gfile.read(4))[0].decode('ascii') == 'HEAD'
        assert struct.unpack('=i', gfile.read(4)) == (264,)
        assert struct.unpack('=i', gfile.read(4)) == (8,)
        size = struct.unpack('=i', gfile.read(4))
    elif size == 65536:     # header size (256) with endian swap
        gformat    = 1
        # the non-native endianness
        endianness = '>' if sys.byteorder == 'little' else '<'
    elif size == 134217728: # header pre-block size (8) with endian swap
        gformat    = 2
        # the non-native endianness
        endianness = '>' if sys.byteorder == 'little' else '<'
        assert struct.unpack(endianness + '4s', gfile.read(4))[0].decode('ascii') == 'HEAD'
        assert struct.unpack(endianness + 'i', gfile.read(4)) == (264,)
        assert struct.unpack(endianness + 'i', gfile.read(4)) == (8,)
        size = struct.unpack(endianness + 'i', gfile.read(4))
    else:
        raise RuntimeError('Encountered unexpected bytes (%d) ' % size + \
                           'at beginning of Gadget file.')
    return gformat, endianness

def read_header(gfile, gformat, endianness):
    '''
    Read and return the header of a single Gadget file.

    Args:
        gfile (file):       The already in binary read mode opened Gadget file or
                            with h5py open HDF5 Gadget file.
        gformat (int):      Gadget file format (either 1, 2, or 3 (HDF5)).
        endianness (str):   The endianness of the file (either native '=' or
                            non-native '<' (little) or '>' (big)). Ignored for
                            HDF5 (format 3) files.

    Returns:
        header (dict):      The header class containting everything of the real
                            header.
    '''
    if gformat not in [1,2,3]:
        raise ValueError('Only formats 1, 2, and 3 (HDF5) are known!')

    header = dict()

    if gformat == 3:
        hattrs = gfile['Header'].attrs

        header['N_part'] = list(hattrs['NumPart_ThisFile'])
        header['mass'] = list(hattrs['MassTable'])
        header['time'] = float(hattrs['Time'])
        header['redshift'] = float(hattrs['Redshift'])
        header['flg_sfr'] = int(hattrs['Flag_Sfr'])
        header['flg_feedback'] = int(hattrs['Flag_Feedback'])
        header['N_part_all'] = list(hattrs['NumPart_Total'].astype(np.uint64) + \
                 (hattrs['NumPart_Total_HighWord'].astype(np.uint64)<<32))
        header['flg_cool'] = int(hattrs['Flag_Cooling'])
        header['N_files'] = int(hattrs['NumFilesPerSnapshot'])
        header['boxsize'] = float(hattrs['BoxSize'])
        header['Omega_m'] = float(hattrs['Omega0'])
        header['Omega_Lambda'] = float(hattrs['OmegaLambda'])
        header['h_0'] = float(hattrs['HubbleParam'])
        header['flg_age'] = int(hattrs['Flag_StellarAge'])
        header['flg_metals'] = int(hattrs['Flag_Metals'])
        header['flg_entropy_instead_u'] = None #int(hattrs['???'])
        header['flg_doubleprecision'] = int(hattrs['Flag_DoublePrecision'])
        header['flg_ic_info'] = int(hattrs['Flag_IC_Info']) \
                if 'Flag_IC_Info' in hattrs else None
        header['lpt_scalingfactor'] = None #float(hattrs['???'])
        header['unused'] = ' '*68  # for consistency
    else:
        gfile.seek(4 if gformat == 1 else 4+8+4+4)

        header['N_part'] = list(struct.unpack(endianness + '6I', gfile.read(6*4)))
        header['mass'] = list(struct.unpack(endianness + '6d', gfile.read(6*8)))
        header['time'], header['redshift'], header['flg_sfr'], \
            header['flg_feedback'] \
                = struct.unpack(endianness + 'd d i i', gfile.read(8+8+4+4))
        header['N_part_all'] = list(struct.unpack(endianness + '6I', gfile.read(6*4)))
        header['flg_cool'], header['N_files'], header['boxsize'], \
            header['Omega_m'], header['Omega_Lambda'], header['h_0'], \
            header['flg_age'], header['flg_metals'], \
            header['flg_entropy_instead_u'], header['flg_doubleprecision'], \
            header['flg_ic_info'], header['lpt_scalingfactor'] \
                = struct.unpack(endianness + 'i i 4d 5i d',
                                gfile.read(2*4+4*8+5*4+8))
        header['unused'] = gfile.read(68).decode('ascii')

        assert struct.unpack(endianness + 'i', gfile.read(4)) == (256,)

    return header

def write_header(gfile, header, gformat, endianness):
    '''
    Write header to the (Gadget-)file gfile with given format and endianness.

    Args:
        gfile (file):       The already in binary write mode opened Gadget file.
        header (dict):      The Gadget header to write.
        gformat (int):      Gadget file format (either 1, 2, or 3 (HDF5)).
        endianness (str):   The endianness of the file (either native '=' or
                            non-native '<' (little) or '>' (big)).
    '''
    if gformat not in [1,2,3]:
        raise ValueError('Only formats 1, 2, and 3 (HDF5) are known!')

    if gformat == 3:
        hat = gfile.create_group('Header').attrs
        hat['NumPart_ThisFile'] = np.array(header['N_part'], dtype=np.int32)
        hat['MassTable'] = np.array(header['mass'], dtype=np.float64)
        hat['Time'] = header['time']
        hat['Redshift'] = header['redshift']
        hat['Flag_Sfr'] = header['flg_sfr']
        hat['Flag_Feedback'] = header['flg_feedback']
        hat['NumPart_Total'] = np.array(header['N_part_all'], dtype=np.uint32)
        hat['NumPart_Total_HighWord'] = np.array(np.array(header['N_part_all'],
                                                           dtype=np.uint64) >> 32,
                                                  np.uint32)
        hat['Flag_Cooling'] = header['flg_cool']
        hat['NumFilesPerSnapshot'] = header['N_files']
        hat['BoxSize'] = header['boxsize']
        hat['Omega0'] = header['Omega_m']
        hat['OmegaLambda'] = header['Omega_Lambda']
        hat['HubbleParam'] = header['h_0']
        hat['Flag_StellarAge'] = header['flg_age']
        hat['Flag_Metals'] = header['flg_metals']
        #hat['???'] = header['flg_entropy_instead_u']
        hat['Flag_DoublePrecision'] = header['flg_doubleprecision']
        if header['flg_ic_info'] is not None:
            hat['Flag_IC_Info'] = header['flg_ic_info']
        #hat['???'] = header['lpt_scalingfactor']
        #header['unused']
    else:
        size = 256
        if gformat == 2:
            _write_format2_leading_block(gfile, 'HEAD', size, endianness)
        gfile.write(struct.pack(endianness + 'i', size))
        start_pos = gfile.tell()

        gfile.write(struct.pack(endianness + '6i', *header['N_part']))
        gfile.write(struct.pack(endianness + '6d', *header['mass']))
        gfile.write(struct.pack(endianness + 'd d i i', header['time'],
                header['redshift'], header['flg_sfr'], header['flg_feedback']))
        gfile.write(struct.pack(endianness + '6i', *header['N_part_all']))
        gfile.write(struct.pack(endianness + 'i i 4d 5i d',
                header['flg_cool'], header['N_files'], header['boxsize'],
                header['Omega_m'], header['Omega_Lambda'], header['h_0'],
                header['flg_age'], header['flg_metals'],
                header['flg_entropy_instead_u'], header['flg_doubleprecision'],
                header['flg_ic_info'], header['lpt_scalingfactor']))
        if isinstance(header['unused'], str):
            gfile.write(header['unused'].encode('ascii'))
        else:
            gfile.write(header['unused'])


        assert gfile.tell() - start_pos == size
        gfile.write(struct.pack(endianness + 'i', size))

class BlockInfo(object):
    '''
    Simple class that holds information of a Gadget file block.

    Args:
        name (str):             The name of the block.
        dtype (str, np.dtype):  The type of the data in form of a
                                numpy.dtype.name, a Gadget style name or a
                                np.dtype.
        dimension (int):        The number of data elements per particle.
        ptypes (list):          A list of booleans telling whether the data is
                                present for a particle type.
        start_pos (int):        The binary start position of the data block on the
                                file or None if it is a HDF5 file.
        size (int):             The size of the data block on the file. None if it
                                is s HDF5 file.
    '''
    _np_type_name = {'LONG':'uint32', 'LLONG':'uint64', 'FLOAT':'float32',
                     'DOUBLE':'float64'}

    def __init__(self, name, dtype, dimension, ptypes, start_pos, size):
        if isinstance(dtype, str) and dtype[-1] == 'N' and dimension < 2:
            raise ValueError('Dimension has to be larger one, if the Gadget ' +
                             'typename ends with "N"!')
        self.name      = name
        self.dtype     = dtype
        if dimension is not None:
            self.dimension = int(dimension)
        else:
            self.dimension = dimension
        self.ptypes    = ptypes
        self.start_pos = start_pos
        self.size      = size

    def is_filled(self, check_type):
        non_none = (self.name, self.start_pos, self.size)
        if check_type:
            non_none += (self.dtype, self.dimension, self.ptypes)
        return not any((prop is None) for prop in non_none)

    @property
    def dtype(self):
        return self._dtype

    @property
    def type_descr(self):
        return self._dtype.kind + str(self._dtype.itemsize)

    @property
    def Gadget_type_name(self):
        Gadget_name = {n:g for g,n
                        in BlockInfo._np_type_name.items()}[self._dtype.name]
        if self.dimension > 1:
            Gadget_name += 'N'
        return Gadget_name

    @dtype.setter
    def dtype(self, dt):
        if dt is None:
            self._dtype = None
            return
        if isinstance(dt, str):
            dt = BlockInfo._np_type_name.get(dt.rstrip('N'), dt)
        dt = np.dtype(dt)
        self._dtype = dt.base

def _read_info(gfile, header, start_pos, block_sizes, info_pos, endianness):
    '''
    Read and return the INFO block from gfile that is positioned at info_pos.

    In practical situations use 'get_block_info' to get the block informations
    from a file. That also works if there is no INFO block.

    Args:
        gfile (file):       The already in binary read mode opened Gadget file.
        header (dict):      The header of the corresponding Gadget file for
                            checking (or to provide hints for deductions).
        start_pos (list):   The start positions of all blocks.
        block_sizes (list): The block sizes of all blocks (pure, the number
                            before the actual block in the file).
        info_pos (int):     The position of the info block (not the leading size,
                            but the actual block).
        endianness (str):   The endianness of the file (either native '=' or
                            non-native '<' (little) or '>' (big)).

    Returns:
        info (dict):        A dictionary with BlockInfo classes as values,
                            containing all information of the blocks.
    '''
    info = {}
    gfile.seek(info_pos)
    for i in range(len(start_pos)):
        if i == 0:                      # HEAD
            pass
        elif start_pos[i] == info_pos:  # INFO block itself
            pass
        else:                           # read info for this block
            name        = struct.unpack(endianness + '4s', gfile.read(4))[0].decode('ascii')
            type_name   = struct.unpack(endianness + '8s', gfile.read(8))[0].strip().decode('ascii')
            dimension   = struct.unpack(endianness + 'i',  gfile.read(4))[0]
            ptypes      = [ n==1 for n in
                    struct.unpack(endianness + '6i', gfile.read(6*4)) ]
            info[name] = BlockInfo(name=name,
                                   dtype=type_name,
                                   dimension=dimension,
                                   ptypes=ptypes,
                                   start_pos=start_pos[i],
                                   size=block_sizes[i])
    return info

def write_info(gfile, info, gformat, endianness):
    '''
    Write the INFO block to the (Gadget-)file gfile with given format and
    endianness.

    Write the Gadget INFO block from the iterable info with BlockInfo elements.
    Here name, type, dimension, and particle types are written. Potentially
    existing blocks in info named HEAD or INFO are skipped.

    Args:
        gfile (file):       The already in binary write mode opened Gadget file.
        info (iterable):    The info to write. The elements have to be BlockInfo
                            classes, which are then going to be written in the
                            order in the iterable, *not* using
                            BlockInfo.start_pos (it is ignored).
        gformat (int):      Gadget file format (either 1 or 2 -- format 3 / HDF5
                            does not have a info block).
        endianness (str):   The endianness of the file (either native '=' or
                            non-native '<' (little) or '>' (big)).
    '''
    if gformat not in [1,2,3]:
        raise ValueError('Only formats 1, 2, and 3 (HDF5) are known!')
    if gformat == 3:
        raise ValueError('HDF5 files (format 3) do not have info blocks!')

    # information of each block is 40 bytes large
    size = len(info) * 40
    if gformat == 2:
        _write_format2_leading_block(gfile, 'INFO', size, endianness)
    gfile.write(struct.pack(endianness + 'i', size))
    start_pos = gfile.tell()

    for block in info:
        gfile.write(struct.pack(endianness + '4s', block.name.encode('ascii')))
        tn = '%-8s' % block.Gadget_type_name
        gfile.write(struct.pack(endianness + '8s', tn.encode('ascii')))
        gfile.write(struct.pack(endianness + 'i',  block.dimension))
        gfile.write(struct.pack(endianness + '6i', *list(map(int,block.ptypes))))

    assert gfile.tell() - start_pos == size
    gfile.write(struct.pack(endianness + 'i', size))

def _fill_block_with_known(block, header):
    '''
    Fill block with the known properties of the block from the config file.
    Also fill the pytes for the MASS block, if not given in the config file.

    Args:
        block (BlockInfo):  The block info class to be filled (if possible).
        header (dict):      The header of the corresponding Gadget file for
                            knowing the number of particle per type.

    Returns:
        type_hint (str):    A hint for the type (e.g. 'float': floating point
                            variable, but unknown whether it is single or double
                            precision).

    Raises:
        RuntimeError:       If the given dtype of the given family is unknown.
    '''
    # read information from config file
    if block.name in config.block_infos:
        info = config.block_infos[block.name]
        block.dimension = info[0]
        if info[1] in ['int', 'uint', 'unsigned', 'float']:
            # type without explicit size -- it's just a hint
            type_hint = info[1]
        else:
            try:
                block.dtype = info[1]
            except TypeError:
                raise RuntimeError('unknown type "%s" for block ' % info[1] +
                                   '"%s"!' % block.name)
            type_hint = None
        if info[2] is None:
            block.dtype = None
        else:
            try:
                fam = list(range(6)) if info[2]=='all' else config.families[info[2]]
                block.ptypes = [(i in fam and header['N_part'][i]>0)
                                    for i in range(6)]
            except KeyError:
                raise RuntimeError('unknown family "%s" for block ' % info[2] +
                                   '"%s"!' % block.name)
    else:
        type_hint = None

    # special treatment of the mass block's ptypes if not given in the config
    # file:
    if block.name == 'MASS' and block.ptypes is None:
        block.ptypes = [(n>0 and m==0) for n,m in
                        zip(header['N_part'],header['mass'])]
        type_hint = 'float'

    return type_hint

def _block_inferring(block, header):
    '''
    Infere those of `dtype`, `dimension`, and `ptypes` that are None for the block
    using its filled properties and combinatorics.

    Args:
        block (BlockInfo):  The block info class to be filled (if possible).
        header (dict):      The header of the corresponding Gadget file to
                            provide hints for deductions.

    Raises:
        RuntimeError:       If the information could not be infered or if the
                            given dtype of the given family is unknown.
                            If, however, there are mutliple combinations ptypes
                            that allow the data type size to be a mutliple of 4
                            bit, do not raise the exception and return the
                            possible combinations.
    '''
    type_hint = _fill_block_with_known(block, header)

    if block.ptypes is None:
        """
        Just do combinatorics: Try all possible combinations of ptypes and collect
        those for which the given block with its size (and its dimension and/or
        dtype, if known) could exist.
        If the dtype is not known, assume its size to be mutliples of 4 (as int32,
        int64, float32, and float64). Furthermore assume that the dimension is not
        larger than 20.
        If there is just one single such combination, we are done. If there is
        more than one, return the combinations, otherwise raise a RuntimeError.
        """
        if block.dimension is not None and block.dtype is not None:
            el_size == block.dimension * block.dtype.itemsize
        else:
            el_size = None
            if block.dimension is not None:
                divisor = 4 * block.dimension
                max_el_size = block.dimension * 8
            elif block.dtype is not None:
                divisor = block.dtype.itemsize
                max_el_size = 20 * block.dtype.itemsize
            else:
                divisor, max_el_size = 4, 20*8
        possible_ptype_combi = []
        for num_ptypes in range(1,7):
            for combi in combinations(list(range(6)), num_ptypes):
                if any(header['N_part'][i] == 0 for i in combi):
                    # avoid N == 0 and multiple combinations with and without
                    # particle types that are not present in this file
                    continue
                N = sum(header['N_part'][i] for i in combi)
                if el_size is not None:
                    if el_size * N == block.size:
                        possible_ptype_combi.append(combi)
                else:
                    element_size = block.size / N
                    if element_size * N == block.size \
                            and element_size % divisor == 0 \
                            and element_size < max_el_size:
                        possible_ptype_combi.append(combi)
        # if the combinatorics yield a definit combination, store it
        if len(possible_ptype_combi) == 1:
            block.ptypes = [(i in possible_ptype_combi[0]) for i in range(6)]
        elif len(possible_ptype_combi) > 1:
            return possible_ptype_combi
        else:
            raise RuntimeError('Could not infere information for block ' +
                               '"%s"!' % block.name)
    # now block.ptypes are known

    if block.dimension is not None and block.dtype is not None:
        return  # all known

    N = sum(header['N_part'][i] for i in range(6) if block.ptypes[i])
    if N == 0:
        # most likely a empty mass block (with masses in header)
        return 'delete'
    element_size = block.size / N
    if block.dtype is not None:
        block.dimension = int(element_size / block.dtype.itemsize)
        return  # all known

    assert block.dtype is None
    # dtype is unknown (and dimension might be as well)...
    if block.dimension is not None:
        itemsize = int(element_size / block.dimension)
        if type_hint == 'int':
            block.dtype = 'i' + str(itemsize)
        elif type_hint in ['uint', 'unsigned']:
            block.dtype = 'u' + str(itemsize)
        else:
            assert type_hint is None or type_hint == 'float'
            # assume float
            block.dtype = 'f' + str(itemsize)
        return  # all known

    # both, dtype and dimension, are unknown...
    if type_hint in ['int', 'uint', 'unsigned']:
        if element_size*8 in [8,16,32,64]:
            block.dtype = type_hint[0] + element_size*8
            block.dimension = 1
        elif element_size % np.dtype(int).itemsize == 0:
            # could be multi-dimensional native
            block.dtype = np.dtype(int)
            block.dimension = int(element_size / block.dtype.itemsize)
        else:
            raise RuntimeError('Could not infere information for block ' +
                               '"%s"!' % block.name)
        assert element_size == block.dimension * block.dtype.itemsize
        return  # all known
    assert type_hint is None or type_hint == 'float'
    # assume float
    block.dtype = 'float64' if header['flg_doubleprecision'] else 'float32'
    block.dimension = int(element_size / block.dtype.itemsize)
    if element_size != block.dimension * block.dtype.itemsize:
        block.dtype = 'float32'
        block.dimension = int(element_size / block.dtype.itemsize)
        if element_size != block.dimension * block.dtype.itemsize:
            block.dtype = None
            block.dimension = None
            raise RuntimeError('Could not infere information for block ' +
                               '"%s"!' % block.name)
    return  # all known

def _infer_info(gfile, header, gformat, endianness, start_pos, block_sizes,
                unclear_blocks):
    '''
    Try to infere block info, if no INFO block is present.

    In practical situations use 'get_block_info' to get the block informations
    from a file.

    Args:
        gfile (file):           The already in binary read mode opened Gadget
                                file.
        header (dict):          The file's header. Used to infer information of
                                the blocks.
        gformat (int):          Gadget file format (either 1 or 2 -- for format 3
                                (HDF5) this function is never needed).
        endianness (str):       The endianness of the file (either native '=' or
                                non-native '<' (little) or '>' (big)).
        start_pos (list):       The start positions of all blocks.
        block_sizes (list):     The block sizes of all blocks (pure, the number
                                before the actual block in the file).
        unclear_blocks (str):   What to do the blocks for which the block info is
                                unclear (cannot be infered). Possible modes are:
                                * exception:    raise an IOError
                                * warning:      print a warning to the stderr
                                * ignore:       guess what

    Returns:
        info (dict):        A dictionary with BlockInfo classes as values,
                            containing all inferred information if the blocks.

    Raises:
        RuntimeError:       If the information could not be infered or if the
                            given dtype of the given family is unknown.
    '''
    if gformat not in [1,2,3]:
        raise ValueError('Only formats 1, 2, and 3 (HDF5) are known!')
    if gformat == 3:
        raise ValueError('HDF5 files (format 3) are not handled here; and this ' +
                         'function is not needed for those.')

    info = {}
    for i in range(len(start_pos)):
        if i == 0:      # HEAD
            continue
        elif block_sizes[i] == (len(block_sizes)-2)*40:
            continue    # INFO
        else:           # actual data block
            if gformat == 1:
                name = config.block_order[i-1] if \
                    i <= len(config.block_order) else '?%03d' % i
            elif gformat == 2:  # can read name
                gfile.seek(start_pos[i]-4-4-8)
                name = struct.unpack(endianness + '4s', gfile.read(4))[0].decode('ascii')
            block = BlockInfo(name=name, dtype=None, dimension=None,
                              ptypes=None, start_pos=start_pos[i],
                              size=block_sizes[i])
            # try to infer type, dimension, and particle types
            combis = _block_inferring(block, header)
            if combis == 'delete':
                continue
            if not block.is_filled(check_type=False):
                assert False    # should never happen
            if not block.is_filled(check_type=True):
                if unclear_blocks == 'exception':
                    raise IOError('Cannot infere info for block "%s"!' %
                            block.name)
                elif unclear_blocks == 'warning':
                    print('WARNING: cannot infere info for ' + \
                                         'block "%s"!\n' % block.name + \
                                         '  possible combinations for the ' + \
                                         'particles types are: ' + \
                                         ', '.join(str(combi) for combi in combis), file=sys.stderr)
                elif unclear_blocks == 'ignore':
                    pass
                else:
                    raise RuntimeError('Unkown block mode ' +
                                       '"%s" is not ' % unclear_blocks +
                                       'understood!')
            info[block.name] = block
    return info

def get_block_info(gfile, gformat, endianness, header, unclear_blocks):
    '''
    Get info (like position in file, if not HDF5) about all blocks in a single
    Gadget file.

    This function utilizes the INFO block of a Gadget file, if (format 1 or 2 and)
    there is a INFO block.
    Otherwise it locates all blocks, does some guessing on the names, if it is a
    format 1 file, and infers the particles for which it is present as well as
    the elment size by combinatorics.

    For HDF5 files, it reads all information available.

    Args:
        gfile (file):           The already in binary read mode opened Gadget
                                file.
        gformat (int):          Gadget file format (either 1, 2, or 3 (HDF5)).
        endianness (str):       The endianness of the file (either native '=' or
                                non-native '<' (little) or '>' (big)).
        header (dict):          The file's header. Used if there is no INFO block
                                and information of the blocks has to inferred.
        unclear_blocks (str):   What to do the blocks for which the block info is
                                unclear (cannot be infered). Possible modes are:
                                * exception:    raise an IOError
                                * warning:      print a warning to the stderr
                                * ignore:       guess what

    Returns:
        info (dict):        A dictionary with BlockInfo classes as values,
                            containing all inferred information if the blocks.

    Raises:
        RuntimeError:       If the information could not be infered or if the
                            given dtype of the given family is unknown.
    '''
    if gformat not in [1,2,3]:
        raise ValueError('Only formats 1, 2, and 3 (HDF5) are known!')

    if gformat == 3:
        info = {}
        for name, group in gfile.items():
            if not name.startswith('PartType'):
                continue
            ptype = int(name[-1])   # name is e.g. 'PartType3'
            # We are accessing the dataset classes, but do not load the actual
            # data! For instance, ds.dtype does not require to load the data (a
            # h5py datastructure is not a np.ndarray!), only actually accessing
            # data within the array (like ds[:] or ds[0]) would load it.
            for ds_name, ds in group.items():
                # for the EAGLE simulations, it happens that there are
                # sub-groups (e.g. for the different element abundancies), hence:
                if not isinstance(ds, h5py.Dataset):
                    # TODO: read the EAGLE elements!
                    continue
                block_name = config.HDF5_to_std_name.get(ds_name, None)
                if block_name is None:
                    block_name = str(ds_name)
                if block_name not in info:
                    info[block_name] = BlockInfo(
                            name=block_name,
                            dtype=ds.dtype.base.name,
                            dimension=ds.shape[1] if len(ds.shape) == 2 else 1,
                            ptypes=[False]*6,
                            start_pos=None,
                            size=None)
                info[block_name].ptypes[ptype] = True
    else:
        # Count blocks and remember start positions.
        # If format 2: Also remember if there is a INFO block and where it is.
        start_pos = []
        block_sizes = []
        if gformat == 2:
            info_pos = None
        gfile.seek(0)
        size = gfile.read(4)
        while size:
            if gformat == 2:
                name = struct.unpack(endianness + '4s', gfile.read(4))[0].decode('ascii')
                gfile.seek(4+4, SEEK_CUR)
                size = gfile.read(4)
            # remember start position and size of block
            start_pos.append(gfile.tell())
            if gformat == 2 and name == 'INFO':
                info_pos = start_pos[-1]
            size, = struct.unpack(endianness + 'i', size)
            block_sizes.append(size)
            # jump over actual block
            gfile.seek(size, SEEK_CUR)
            assert struct.unpack(endianness + 'i', gfile.read(4)) == (size,)
            # try to read size of next block
            size = gfile.read(4)

        if gformat == 1 and block_sizes[-1] == (len(block_sizes)-2)*40:
            # If there is a INFO block, it has to be at the end and of size
            # (#blocks-2)*40.
            info = _read_info(gfile, header, start_pos, block_sizes,
                              start_pos[-1], endianness)
        elif gformat == 2 and info_pos:
            info = _read_info(gfile, header, start_pos, block_sizes, info_pos,
                              endianness)

        else:
            # no info block -> infer as much as possible (using block names in
            # format 2 files)
            info = _infer_info(gfile, header, gformat, endianness, start_pos,
                               block_sizes, unclear_blocks=unclear_blocks)

    if 'MASS' not in info:
        # just put it after ID block as usual
        if gformat==3:
            pos = None
        else:
            pos = info['ID  '].start_pos + info['ID  '].size + \
                    (info['ID  '].start_pos -
                            info['VEL '].start_pos + info['VEL '].size)
        info['MASS'] = BlockInfo(
                name='MASS',
                dtype='float64' if header['flg_doubleprecision']
                            else 'float32',
                dimension=1,
                ptypes=[False]*6,
                start_pos=pos,
                size=0)

    return info

def _write_format2_leading_block(gfile, name, size, endianness):
    '''Little helper function with speaking name, that writes the small leading
    blocks for format 2 Gadget files.'''
    gfile.write(struct.pack(endianness + ' i 4s i i', 8, name.encode('ascii'), size+8, 8))

def write_block(gfile, block_name, data, gformat, endianness='=',
                gad_units=None):
    '''
    Write a block to the (Gadget-)file gfile with given format and endianness.

    Args:
        gfile (file):       The already in binary write mode opened Gadget file.
        block_name (str):   The block name for the block to write.
        data (...):         The data to write. A UnitArr (or simplye a
                            numpy.array) for regular blocks, a header dict for
                            the HEAD block and an iterable with BlockInfo classes
                            as elements for the INFO block.
        gformat (int):      Gadget file format (either 1 or 2 -- format 3 (HDF5)
                            should be written explicitly).
        endianness (str):   The endianness of the file (either native '=' or
                            non-native '<' (little) or '>' (big)).
        gad_units (dict):   The basic gadget units. If None, take the units
                            defined in the `gadget.cfg`.
    '''
    if gformat not in [1,2,3]:
        raise ValueError('Only formats 1, 2, and 3 (HDF5) are known!')
    if gformat == 3:
        raise NotImplementedError('Writing HDF5 blocks shall not be done by ' + \
                                  'this function!')

    if gad_units is None:
        gad_units = config.default_gadget_units

    # warn if the data has units but these are not the default units
    if hasattr(data,'unit') and block_name in config._block_units \
            and data.unit != config.get_block_units(block_name,gad_units):
        warn_msg = "Your are writing the block '%s' in other " % block_name + \
                   "units ('%s') than the default units " % data.unit + \
                   "('%s')!" % config.get_block_units(block_name,gad_units)
        print('WARNING:', warn_msg, file=sys.stderr)
    # reduce data to the numpy array
    if not isinstance(data, np.ndarray):
        data = np.array(data)

    size = data.nbytes
    if gformat == 2:
        _write_format2_leading_block(gfile, block_name, size, endianness)
    gfile.write(struct.pack(endianness + 'i', size))
    start_pos = gfile.tell()

    data.tofile(gfile)

    assert gfile.tell() - start_pos == size
    gfile.write(struct.pack(endianness + 'i', size))

